// ===================================================================
// $Id: $
//
// hdrloader.h
//
// Class: CHDRImage, CEnvironmentMap
//
//  The declaration of the class for reading HDR image and environment maps
//
// Initial coding by Vlastimil Havran, 2013
// Some routines by Igor Kravtchenko, 2002 (reading HDR image)

#include <cmath>
#include <cstdio>
#include <cstdlib>
#include <string.h>

#include "configg.h"
#include "basstr.h"
#include "basmath.h"
#include "colorfs.h"
#include "hdrloader.h"
#include "lightenv.h"
#include "readpng.h"
#include "rgbe.h"
#include "tgaimg.h"
#include "TAlloc.h"
#include "AuxFuncs.h"
#include "AuxRND.h"
#include "tangles.h"

// OpenEXR format
#ifdef _ALLOW_OPENEXR_
#include <ImfOutputFile.h>
#include <ImfChannelList.h>
#include <ImfStringAttribute.h>
#include <ImfIO.h>
//#include <ImfInputFile.h>
//#include <ImfVersion.h>
//#include <ImfRgbaFile.h>
//#include <ImfStringAttribute.h>
//#include <ImfMatrixAttribute.h>
//#include <ImfArray.h>
using namespace std;
using namespace Imf;
//using namespace Imath;
#endif // _ALLOW_OPENEXR_


typedef unsigned char RGBE[4];
#define R			0
#define G			1
#define B			2
#define E			3

#define  MINELEN	8	  // minimum scanline length for encoding
#define  MAXELEN	0x7fff	  // maximum scanline length for encoding


float
convertComponent(int expo, int val)
{
  float v = val / 256.0f;
  float d = (float) pow(2, expo);
  return v * d;
}

void
workOnRGBE(RGBE *scan, int len, float *cols)
{
  while (len-- > 0) {
    int expo = scan[0][E] - 128;
    float s = 0.f;
    cols[0] = convertComponent(expo, scan[0][R]);
    s += cols[0];
    cols[1] = convertComponent(expo, scan[0][G]);
    s += cols[1];
    cols[2] = convertComponent(expo, scan[0][B]);
    s += cols[2];
    // summing to floats requires that non-zero
    // values are used as luminances. Completely
    // black value does not allow to use binary search
    // when doing importance sampling
    const float threshold = 3e-4f;
    if (s < threshold) {
      const float minValue = 0.3333*threshold;
      cols[0] = minValue;
      cols[1] = minValue;
      cols[2] = minValue;
    }
    cols += 3;
    scan++;
  }
  return;
}

bool
oldDecrunch(RGBE *scanline, int len, FILE *file)
{
  int i;
  int rshift = 0;
  
  while (len > 0) {
    scanline[0][R] = fgetc(file);
    scanline[0][G] = fgetc(file);
    scanline[0][B] = fgetc(file);
    scanline[0][E] = fgetc(file);
    if (feof(file))
      return false;
    
    if (scanline[0][R] == 1 &&
	scanline[0][G] == 1 &&
	scanline[0][B] == 1) {
      for (i = scanline[0][E] << rshift; i > 0; i--) {
	memcpy(&scanline[0][0], &scanline[-1][0], 4);
	scanline++;
	len--;
      }
      rshift += 8;
    }
    else {
      scanline++;
      len--;
      rshift = 0;
    }
  }
  return true;
}

bool
decrunch(RGBE *scanline, int len, FILE *file)
{
  int  i, j;
  
  if (len < MINELEN || len > MAXELEN)
    return oldDecrunch(scanline, len, file);
  
  i = fgetc(file);
  if (i != 2) {
    fseek(file, -1, SEEK_CUR);
    return oldDecrunch(scanline, len, file);
  }
  
  scanline[0][G] = fgetc(file);
  scanline[0][B] = fgetc(file);
  i = fgetc(file);
  
  if (scanline[0][G] != 2 || scanline[0][B] & 128) {
    scanline[0][R] = 2;
    scanline[0][E] = i;
    return oldDecrunch(scanline + 1, len - 1, file);
  }

  // read each component
  for (i = 0; i < 4; i++) {
    for (j = 0; j < len; ) {
      unsigned char code = fgetc(file);
      if (code > 128) { // run
	code &= 127;
	unsigned char val = fgetc(file);
	while (code--)
	  scanline[j++][i] = val;
      }
      else  {	// non-run
	while(code--)
	  scanline[j++][i] = fgetc(file);
      }
    }
  }

  return feof(file) ? false : true;
}

CHDRImage::CHDRImage(const CHDRImage &src)
{
  this->cols = 0;
  this->allocate(src.width, src.height, src.typeHDR);
  this->filenameRead = src.filenameRead; // if read, then filename
  this->gname = src.gname; // the name without extension and path
  this->copyDataFromArray(width, height, 3, src.cols, 0, 0, false);
}

CHDRImage::CHDRImage(const CHDRImage &src, int pseudoColorTechnique,
		     bool copyImage, int widthBargraph, int borderBargraph)
{
  this->cols = 0;
  this->allocate(src.getWidth() + widthBargraph + borderBargraph, src.getHeight(), src.typeHDR);
  int w2 = src.getWidth();
  for (int y = 0; y < height; y++) {
    if (copyImage) {
      for (int x = 0; x < w2; x++) {
	int wi = (w2*y+x)*3; // source
	int wo = (width*y+x)*3; // target
	cols[wo+0] = src.cols[wi+0];
	cols[wo+1] = src.cols[wi+1];
	cols[wo+2] = src.cols[wi+2];
	//setRGB_XY(x, y, imgBTF[i], imgBTF[i+1], imgBTF[i+2]);
      } // for x
    }
    if (1) {
      // copy the bargraph values to the right
      float RGB[3];
      float value = (float)(height-y)/(float)height; // get the color
      GetPseudoColor(pseudoColorTechnique, value, 0.0, 1.0, RGB);
      for (int x = w2; x < w2 + borderBargraph; x++) {
	setRGB_XY(x, y, 0, 0, 0); // make it black
      }
      for (int x = w2 + borderBargraph; x < width; x++) {
	setRGB_XY(x, y, RGB[0], RGB[1], RGB[2]);	
      }
    }
  } // for y
  return;
}


void
CHDRImage::Multiply(float coeff, int x1, int y1, int x2, int y2)
{
  if (x2 < 0)
    x2 = width;
  if (y2 < 0)
    y2 = height;
  
  // Copy the data from PNG representation
  for (int y=y1; y < y2; y++) {
    for (int x=x1; x < x2; x++) {
      int ww = (width * y + x)*3;
      for (int isp=0; isp < 3; isp++) {
	// set it to the output format as required by RGBE_WRITE
	cols[ww+isp] *= coeff;
      }
    } // for jcol
  } // for irow
}

void
CHDRImage::AddOffset(float coeff, int x1, int y1, int x2, int y2)
{
  if (x2 < 0)
    x2 = width;
  if (y2 < 0)
    y2 = height;
  
  // Copy the data from PNG representation
  for (int y=y1; y < y2; y++) {
    for (int x=x1; x < x2; x++) {
      int ww = (width * y + x)*3;
      for (int isp=0; isp < 3; isp++) {
	// set it to the output format as required by RGBE_WRITE
	cols[ww+isp] += coeff;
      }
    } // for jcol
  } // for irow
  return;
}

// This allows limit if from the bottom
void
CHDRImage::Maximize(float coeff, int x1, int y1, int x2, int y2)
{
  if (x2 < 0)
    x2 = width;
  if (y2 < 0)
    y2 = height;
  
  // Copy the data from PNG representation
  for (int y=y1; y < y2; y++) {
    for (int x=x1; x < x2; x++) {
      int ww = (width * y + x)*3;
      for (int isp=0; isp < 3; isp++) {
	// set it to the output format as required by RGBE_WRITE
	if (cols[ww+isp] < coeff)
	  cols[ww+isp] = coeff;
      }
    } // for jcol
  } // for irow
}

 // This allows limit if from the top
void
CHDRImage::Minimize(float coeff, int x1, int y1, int x2, int y2)
{
  if (x2 < 0)
    x2 = width;
  if (y2 < 0)
    y2 = height;
  
  // Copy the data from PNG representation
  for (int y=y1; y < y2; y++) {
    for (int x=x1; x < x2; x++) {
      int ww = (width * y + x)*3;
      for (int isp=0; isp < 3; isp++) {
	// set it to the output format as required by RGBE_WRITE
	if (cols[ww+isp] > coeff)
	  cols[ww+isp] = coeff;
      }
    } // for jcol
  } // for irow
  return;
}

// This makes absolute value of the image
void
CHDRImage::Fabs(int x1, int y1, int x2, int y2)
{
  if (x2 < 0)
    x2 = width;
  if (y2 < 0)
    y2 = height;
  
  // Copy the data from PNG representation
  for (int y=y1; y < y2; y++) {
    for (int x=x1; x < x2; x++) {
      int ww = (width * y + x)*3;
      for (int isp=0; isp < 3; isp++) {
	// set it to the output format as required by RGBE_WRITE
	if (cols[ww+isp] < 0)
	  cols[ww+isp] = -cols[ww+isp];
      }
    } // for jcol
  } // for irow
  return;
}

void
CHDRImage::MakeGray(int x1, int y1, int x2, int y2)
{
  // Copy the data from PNG representation
  for (int y=y1; y < y2; y++) {
    for (int x=x1; x < x2; x++) {
      int ww = (width * y + x)*3;
      float lum = luminance(cols+ww);
      // turn it to gray, whole image
      cols[ww+0] = lum;
      cols[ww+1] = lum;
      cols[ww+2] = lum;
    } // for jcol
  } // for irow
  return;
}

void
CHDRImage::MakeWholeGray()
{
  // Copy the data from PNG representation
  for (int y=0; y < height; y++) {
    for (int x=0; x < width; x++) {
      int ww = (width * y + x)*3;
      float lum = luminance(cols+ww);
      // turn it to gray, whole image
      cols[ww+0] = lum;
      cols[ww+1] = lum;
      cols[ww+2] = lum;
    } // for jcol
  } // for irow
  return;
}

bool
CHDRImage::loadHDR(const char *fileName)
{
  int i;
  char str[200];
  FILE *file;

  file = fopen(fileName, "rb");
  if (!file)
    return false; // error

  //int tt = fread(str, 10, 1, file);
  char *tt = fgets(str, 14, file);
  if ((memcmp(str, "#?RADIANCE", 10)!=0)&&(memcmp(str, "#?RGBE", 6)!=0)) {
    fclose(file);
    return false; // error
  }

  fseek(file, 1, SEEK_CUR);

  char cmd[200];
  i = 0;
  char c = 0, oldc;
  while(true) {
    oldc = c;
    c = fgetc(file);
    if (c == 0xa && oldc == 0xa)
      break;
    cmd[i++] = c;
  }

  char reso[200];
  i = 0;
  while(true) {
    c = fgetc(file);
    reso[i++] = c;
    if (c == 0xa)
      break;
  }

  int w = 0, h = 0;
  if (sscanf(reso, "-Y %d +X %d", &h, &w) != 2) {
    fclose(file);
    return false; // error
  }
  
  this->width = w;
  this->height = h;  
  printf("Reading HDR image '%s'\n", fileName);
  cout << "HDR loader width = " << width << " height = " << height << endl;
  allocate(width, height, true);
  assert(this->cols);
  float *localcols = cols;
  //cout << "cols = " << cols << endl;
  
  RGBE *scanline = new RGBE[w * 4];
  assert(scanline);
  if (!scanline) {
    fclose(file);
    return false; // error
  }
  
  // convert image 
  for (int y = h - 1; y >= 0; y--) {
    if (decrunch(scanline, w, file) == false)
      break;
    workOnRGBE(scanline, w, localcols);
    localcols += w * 3;
  }
  
  delete [] scanline;
  fclose(file);

  extractName(fileName);

  //cout << "cols = " << cols << endl;
  return true; // OK
}

bool
CHDRImage::saveHDR(char *fileName, bool verbose)
{
  string fileNameReal;
#ifdef _ALLOW_OPENEXR_
  // change hdr to EXR
  {
    fileNameReal = ChangeFilenameExtension(string(fileName), string("exr"));
    if (verbose)
      cout << "Writing to EXR file: " <<  fileNameReal << endl;
    Imf::Header header(width, height);
    header.insert("comments", Imf::StringAttribute("Generated by C++"));
    Imf::ChannelList &channels = header.channels();
    channels.insert("R", Imf::Channel(Imf::FLOAT));
    channels.insert("G", Imf::Channel(Imf::FLOAT));
    channels.insert("B", Imf::Channel(Imf::FLOAT));
    Imf::FrameBuffer frameBuffer;
    size_t compStride = sizeof(float);
    size_t pixelStride = 3 * compStride;
    size_t rowStride = pixelStride * width;
    char *ptr = reinterpret_cast<char *>(cols);
    frameBuffer.insert("R", Imf::Slice(Imf::FLOAT, ptr, pixelStride, rowStride));
    ptr += compStride;
    frameBuffer.insert("G", Imf::Slice(Imf::FLOAT, ptr, pixelStride, rowStride));
    ptr += compStride;
    frameBuffer.insert("B", Imf::Slice(Imf::FLOAT, ptr, pixelStride, rowStride));
    Imf::OutputFile file(fileNameReal.c_str(), header);
    file.setFrameBuffer(frameBuffer);
    file.writePixels(height); // how many rows
    printf("Writing to HDR file %s: %d\n", fileNameReal.c_str(), 0);
  }
#else // _ALLOW_OPENEXR_
  // saving in HDR format by Greg Ward
  fileNameReal = ChangeFilenameExtension(string(fileName), string("hdr"));
  int planesCnt = 3;
  float ***imagefloat = allocation3(0, planesCnt, 0, height, 0, width);
  // Copy the data from PNG representation
  for (int y=0; y < height; y++) {
    for (int x=0; x < width; x++) {
      int ww = (width * y + x)*3;
      for (int isp=0; isp < planesCnt; isp++) {
	// set it to the output format as required by RGBE_WRITE
	imagefloat[isp][y][x] = cols[ww+isp];
      }
    } // for jcol
  } // for irow
  
  int statusWrite = write_RGBE_file((char*)(fileNameReal.c_str()), imagefloat,
				    planesCnt, height, width);
  printf("Writing to HDR file %s: %d\n", fileNameReal.c_str(), statusWrite);
  if (statusWrite < 0) {
    return true; // ERROR
  }
  // deallocate the data needed for writing
  freemem3(imagefloat, 0, planesCnt, 0, height, 0, width);
#endif // _ALLOW_OPENEXR_
  return false; // OK
}

bool
CHDRImage::loadPNG(const char *fileName)
{
  // First, read header
  int planesCnt;
  int readStatus = readPNGheader(fileName, &planesCnt, &height, &width);
  if (readStatus < 0) {
    width = height = 0;
    printf("Wrong environment map file '%s'!!! Exiting ...\n", fileName);
    return false; // error
  }

  // for PNG
  float ***imagefloat = 0; 
  // allocation of PNG env. image
  imagefloat = allocation3(0, planesCnt, 0, height, 0, width);
  assert(imagefloat);
  assert(planesCnt == 3);
  
  // reading PNG env. map file
  readStatus = readPNGfile3D(fileName, imagefloat, planesCnt, height, width, 1);
  printf("Reading PNG image '%s'(%d)\n", fileName, readStatus);
  cout << "PNG loader width = " << width << " height = " << height << endl;
  allocate(width, height, false);
  assert(cols);

  // Copy the data from PNG representation
  for (int y=0; y < height; y++) {
    for (int x=0; x < width; x++) {
      int ww = (width * y + x)*3;
      float s = 0.f;
      for (int isp=0; isp < planesCnt; isp++) {
	float val = imagefloat[isp][y][x];
	cols[ww+isp] = val;
	s += val;
      }
#if 1
      // summing to floats requires that non-zero
      // values are used as luminances. Completely
      // black value does not allow to use binary search
      // when doing importance sampling
      const float threshold = 1e-4f;
      if (s < threshold) {
	const float minValue = 0.3333f*threshold;
	cols[ww+0] = minValue;
	cols[ww+1] = minValue;
	cols[ww+2] = minValue;
      }
#endif
    } // for jcol
  } // for irow

  // deallocate the data as read by PNG reader
  freemem3(imagefloat, 0, planesCnt, 0, height, 0, width);

  extractName(fileName);
  return true; // OK
}

void
CHDRImage::extractName(const char *fileName)
{
  // makes just a single name
  filenameRead.assign(fileName);
  string gname2; // remove 'prepend' path
  RemovePathFromFilename(filenameRead, gname2);
  // store basic name to 'gname'
  ChangeFilenameExtension(gname2, string(""), gname);
  //cout << "filenameRead = " << filenameRead << " gname = " << gname << endl;
  return;
}

// shifts circularly the image to left or right
void
CHDRImage::horizontallyShift(int pixelOffset)
{
  int planesCnt = 3;
  float ***rowaux = allocation3(0, planesCnt, 0, 1, 0, width);

  // Copy the data from PNG representation
  for (int y=0; y < height; y++) {
    // copy the data to a temporary buffer
    for (int x=0; x < width; x++) {
      int ww = (width * y + x)*3;
      for (int isp=0; isp < planesCnt; isp++) {
	// set it to the output format as required by RGBE_WRITE
	rowaux[isp][0][(x+pixelOffset)%width] = cols[ww+isp];
      }
    } // for whole row
    // and save it back to the same line
    for (int x=0; x < width; x++) {
      int ww = (width * y + x)*3;
      for (int isp=0; isp < planesCnt; isp++) {
	cols[ww+isp] = rowaux[isp][0][x];
      }
    } // for whole row
  } // for irow

  // deallocate the data needed for writing
  freemem3(rowaux, 0, planesCnt, 0, 1, 0, width);
    
  return;
}


bool
CHDRImage::savePNG(const char *fileName)
{
  int planesCnt = 3;
  float ***imagefloat = allocation3(0, planesCnt, 0, height, 0, width);

  // Copy the data from PNG representation
  for (int y=0; y < height; y++) {
    for (int x=0; x < width; x++) {
      int ww = (width * y + x)*3;
      for (int isp=0; isp < planesCnt; isp++) {
	// set it to the output format as required by RGBE_WRITE
	imagefloat[isp][y][x] = cols[ww+isp];
      }
    } // for jcol
  } // for irow
  
  int statusWrite = writePNGfile((char*)fileName, imagefloat, planesCnt, height, width, 0, 1);
  printf("HDR write %s: %d\n", fileName, statusWrite);
  if (statusWrite < 0) {
    return true; // ERROR
  }

  // deallocate the data needed for writing
  freemem3(imagefloat, 0, planesCnt, 0, height, 0, width);

  return false; // OK
}

void
CHDRImage::makeMarkRGB(int mx, int my, EMarkType typeMark, int marksize, float mult)
{
  int x, y;
  
  float r, g, b;
  if (typeMark == EE_BRDFSamples) {
    r = g = b = 255.0; // white
  }
  else {
    // ENVMAP samples
    r = 255.0; g = 5.0; b = 5.0; // red
  }
  // Make mark with correct color
  r *= mult;
  g *= mult;
  b *= mult;

  if (typeMark == EE_BRDFSamples) {
    // the horizontal direction
    if ((my >= 0) && (my < height)) {
      for (x = mx - marksize; x <= mx + marksize; x++) {
	if ((x < 0) || (x >= width))
	  continue;
	setRGB_XY(x, my, r, g, b);
      }
    }
    if ((mx >= 0) && (mx < width)) {
      for (y = my - marksize; y <= my + marksize; y++) {
	if ((y < 0) || (y >= height))
	  continue;
	setRGB_XY(mx, y, r, g, b);
      }
    }
  }
  else { // environment map samples
    int t;
    // from left bottom to right top corner
    for (t = - marksize; t <= marksize; t++) {
      x = mx + t;
      if ((x < 0) || (x >= width))
	continue;
      y = my + t;
      if ((y < 0) || (y >= height))
	continue;
      setRGB_XY(x, y, r, g, b);
    } // for t
    // from left top to right bottom corner
    for (t = - marksize; t <= marksize; t++) {
      x = mx + t;
      if ((x < 0) || (x >= width))
	continue;
      y = my - t;
      if ((y < 0) || (y >= height))
	continue;
      setRGB_XY(x, y, r, g, b);
    } // for t
  }

  return; // mark finished
}

// Trivial algorithm to render a line for purposes of visualization
void
CHDRImage::drawLine(int x1, int y1, int x2, int y2,
		    float red, float green, float blue)
{
  if (fabs(float(x1 - x2)) > fabs(float(y1 - y2))) {
    // the loop along the x-axis
    int deltax = (x2 - x1) > 0 ? 1: -1;
    for (int x = x1; x < x2; x++) {
      int y = y1 + (y2 - y1) * (x - x1) / (x2 - x1);
      setRGB_XY(x, y, red, green, blue);
    }
  }
  else {
    // the loop along the y-axis
    int deltay = (y2 - y1) > 0 ? 1 : -1;
    for (int y = y1; y < y2; y++) {
      int x = x1 + (x2 - x1) * (y - y1) / (y2 - y1);
      setRGB_XY(x, y, red, green, blue);
    }
  }
  
  return;
}

// the whole image is filled in by the same color
void
CHDRImage::setConstant(float r, float g, float b)
{
  // Copy the data from PNG representation
  for (int y=0; y < height; y++) {
    for (int x=0; x < width; x++) {
      int ww = (width * y + x)*3;
      // set it to the output format as required by RGBE_WRITE
      cols[ww+0] = r;
      cols[ww+1] = g;
      cols[ww+2] = b;
    } // for jcol
  } // for irow
  return;
}

int
CHDRImage::setsPseudocolor(CHDRImage &i2, int pseudoColorTechnique,
			   float lowerBound, float upperBound) const
{
  int w2 = Min(i2.width, width);
  int h2 = Min(i2.height, height);
  bool differ = false;
  if ((i2.width != width)||
      (i2.height != height)) {
    cout << "The size of image differ, selecting smaller sizes: "
	 << w2 << " x " << h2 << endl;
    differ = true;
  }
  // Copy the data from PNG representation
  float RGB[3];
  for (int y=0; y < h2; y++) {
    for (int x=0; x < w2; x++) {
      int ww = (width * y + x)*3;
      float lum1 = luminance(&(cols[ww+0])); // of this image
      GetPseudoColor(pseudoColorTechnique, lum1, lowerBound, upperBound, RGB);
      i2.setValueChannel(x, y, 0, RGB[0]);
      i2.setValueChannel(x, y, 1, RGB[1]);
      i2.setValueChannel(x, y, 2, RGB[2]);
    } // for jcol
  } // for irow
  return false; // done
}

// Substructs two images and prints possibly the output, that is created
// by this function
CHDRImage*
CHDRImage::Substruct(CHDRImage &i2, CHDRImage *i3, bool computeRMSE, int verbose) const
{
  int w2 = Min(i2.width, width);
  if (i3)
    w2 = Min(w2, i3->width);
  
  int h2 = Min(i2.height, height);
  if (i3)
    h2 = Min(h2, i3->height);

  bool differ = false;
  if ((i2.width != width)||
      (i2.height != height)) {
    cout << "The size of image differ, selecting smaller sizes: "
	 << w2 << " x " << h2 << endl;
    differ = true;
  }
  CHDRImage* ires = 0;
  if (i3 == 0) {
    ires = new CHDRImage();
    ires->allocate(w2, h2);
  }
  else // we got the image
    ires = i3;
  
  long countPixels = w2 * h2;
  // the number of all different pixels
  long countDiffPixels = 0L;

  // the sum and square sum
  long double sumRGB = 0.0;
  long double sum2RGB = 0.0;
  long double sumLum = 0.0;
  long double sum2Lum = 0.0;
  // Copy the data from PNG representation
  for (int y=0; y < h2; y++) {
    long double sumRGBForScanline = 0;
    long double sumLumForScanline = 0;
    long double sumRGBSquaredForScanline = 0;
    long double sumLumSquaredForScanline = 0;
    for (int x=0; x < w2; x++) {
      int ww = (width * y + x)*3;
      int ww2 = (i2.width * y + x)*3;
      int wwres = (w2 * y + x)*3;
      float diffr = (cols[ww+0] - i2.cols[ww2+0]);
      float diffg = (cols[ww+1] - i2.cols[ww2+1]);
      float diffb = (cols[ww+2] - i2.cols[ww2+2]);
      ires->cols[wwres+0] = diffr;
      ires->cols[wwres+1] = diffg;
      ires->cols[wwres+2] = diffb;
      //------------------------
      if (computeRMSE) {
	float lum1 = luminance(&(cols[ww+0]));
	float lum2 = luminance(&(i2.cols[ww2+0]));
	sumRGBForScanline += fabs(diffr)+fabs(diffg)+fabs(diffb);
	sumRGBSquaredForScanline += diffr*diffr+diffg*diffg+diffb*diffb;
	sumLumForScanline += fabs(lum1-lum2);
	sumLumSquaredForScanline += sqr(lum1-lum2);
	const float epsilonLum = 1e-20;      
	if ((lum1 - lum2) > epsilonLum)
	  countDiffPixels++;
      }
    } // for jcol
    if (computeRMSE) {
      sumRGB += sumRGBForScanline;
      sum2RGB += sumRGBSquaredForScanline;
      sumLum += sumLumForScanline;
      sum2Lum += sumLumSquaredForScanline;
    }
  } // for irow

  if (computeRMSE) {
    // compute root-mean-square error across the whole image
    long double RMSErrorLum = sqrt(sum2Lum / (long double)(countPixels));
    long double RMSErrorRGB = sqrt(sum2RGB / (long double)(countPixels));
    
    // compute standard deviation from all different pixels of the image
    long double devLum = 1.0 / (long double)countPixels * sum2Lum - sumLum * sumLum /
      ( (long double)countPixels * (long double)countPixels);
    long double devRGB = 1.0 / (long double)countPixels * sum2RGB - sumRGB * sumRGB /
      ((long double)countPixels / (long double)countPixels)*9.f;
    if (devRGB > 0.0)
      devRGB = sqrt(devRGB);
    else
      devRGB = 0.f;
    if (devLum > 0.0)
      devLum = sqrt(devLum);
    else
      devLum = 0.f;
    
    long double defChangedPixelsLum = 0.0;
    long double RMSErrorChangedPixelsLum = 0.0;
    long double defChangedPixelsRGB = 0.0;
    long double RMSErrorChangedPixelsRGB = 0.0;
    
    // compute RMSE and the standard deviation from all the pixels changed
    if (countDiffPixels) {
      defChangedPixelsLum = 1.0 / (long double)countDiffPixels * sum2Lum -
	sumLum * sumLum / (long double)countDiffPixels /
	(long double)countDiffPixels;
      if (defChangedPixelsLum > 0.f)
	defChangedPixelsLum = sqrt(defChangedPixelsLum);
      else
	defChangedPixelsLum = 0;
      defChangedPixelsRGB = 1.0 / (long double)countDiffPixels * sum2RGB -
	sumRGB * sumRGB / (9*(long double)countDiffPixels *
			   (long double)countDiffPixels);
      if (defChangedPixelsRGB > 0.f)
	defChangedPixelsRGB = sqrt(defChangedPixelsRGB);
      else
	defChangedPixelsRGB = 0;
      
      RMSErrorChangedPixelsLum = sqrt(sum2Lum / (long double)(countDiffPixels));
      RMSErrorChangedPixelsRGB = sqrt(sum2RGB / (long double)(9*countDiffPixels));
    }

    // computing PSNR - peak signal-to-reconstruced image measure
    long double PSNR = 20 * log10(255.0 / RMSErrorLum);
    long double PSNR_RGB = 20 * log10(255.0 / RMSErrorRGB);
    if (verbose) {
      cout << "Comparing two images by RMSE (root-mean-squared) error"
	   << endl;
      cout << "Percentage of changed pixels is "
	   << 100.0f * countDiffPixels / countPixels << " [%] ("
	   << countDiffPixels << " from cntAllPixels = " << countPixels
	   << ")" << endl;
      cout << "For all pixels RMSE error[RGB,LUM] = "
	   << RMSErrorRGB * 100.0f << " " << RMSErrorLum * 100.0f << " [%] " << endl; 
      cout << "For only changed pixels RMSE error = "
	   << RMSErrorChangedPixelsRGB * 100.0f << " " << RMSErrorChangedPixelsLum * 100.0f << " [%] " << endl; 
      cout << "Standard deviation across whole image is = " << devRGB << " " << devLum << endl;
      cout << "Standard deviation for changed pixels only image is = "
	   << defChangedPixelsRGB << " " << defChangedPixelsLum << endl;
      cout << "For all pixels PSNR = " << PSNR_RGB << " " << PSNR << " [decibels] " << endl;
      cout << "Main result from comparing two images = " << RMSErrorLum << endl;
    }
  }
  return ires; // the same size .. false, different size .. true
}

// Compare two images and prints possibly the output
bool
CHDRImage::CompareTo(CHDRImage &i2, int verbose) const
{
  int w2 = Min( i2.width, width);
  int h2 = Min(i2.height, height);
  bool differ = false;
  if ((i2.width != width)||
      (i2.height != height)) {
    cout << "The size of image differ, selecting smaller sizes: "
	 << w2 << " x " << h2 << endl;
    differ = true;
  }
  
  long countPixels = w2 * h2;
  // the number of all different pixels
  long countDiffPixels = 0L;

  // the sum and square sum
  long double sumRGB = 0.0;
  long double sum2RGB = 0.0;
  long double sumLum = 0.0;
  long double sum2Lum = 0.0;
  // Copy the data from PNG representation
  for (int y=0; y < h2; y++) {
    long double sumRGBForScanline = 0;
    long double sumLumForScanline = 0;
    long double sumRGBSquaredForScanline = 0;
    long double sumLumSquaredForScanline = 0;
    for (int x=0; x < w2; x++) {
      int ww = (width * y + x)*3;
      int ww2 = (i2.width * y + x)*3;
      float diffr = (cols[ww+0] - i2.cols[ww2+0]);
      float diffg = (cols[ww+1] - i2.cols[ww2+1]);
      float diffb = (cols[ww+2] - i2.cols[ww2+2]);
      float lum1 = luminance(&(cols[ww+0]));
      float lum2 = luminance(&(i2.cols[ww2+0]));
      sumRGBForScanline += fabs(diffr)+fabs(diffg)+fabs(diffb);
      sumRGBSquaredForScanline += diffr*diffr+diffg*diffg+diffb*diffb;
      sumLumForScanline += fabs(lum1-lum2);
      sumLumSquaredForScanline += sqr(lum1-lum2);
      const float epsilonLum = 1e-20;      
      if ((lum1 - lum2) > epsilonLum)
	countDiffPixels++;
    } // for jcol
    sumRGB += sumRGBForScanline;
    sum2RGB += sumRGBSquaredForScanline;
    sumLum += sumLumForScanline;
    sum2Lum += sumLumSquaredForScanline;
  } // for irow
  
  // compute root-mean-square error across the whole image
  long double RMSErrorLum = sqrt(sum2Lum / (long double)(countPixels));
  long double RMSErrorRGB = sqrt(sum2RGB / (long double)(3*countPixels));

  // compute standard deviation from all different pixels of the image
  long double devLum = 1.0 / (long double)countPixels * sum2Lum - sumLum * sumLum /
    ( (long double)countPixels * (long double)countPixels);
  long double devRGB = 1.0 / (long double)countPixels * sum2RGB - sumRGB * sumRGB /
    ((long double)countPixels / (long double)countPixels)*9.f;
  if (devRGB > 0.0)
    devRGB = sqrt(devRGB);
  else
    devRGB = 0.f;
  if (devLum > 0.0)
    devLum = sqrt(devLum);
  else
    devLum = 0.f;
    
  long double defChangedPixelsLum = 0.0;
  long double RMSErrorChangedPixelsLum = 0.0;
  long double defChangedPixelsRGB = 0.0;
  long double RMSErrorChangedPixelsRGB = 0.0;
  
  // compute RMSE and the standard deviation from all the pixels changed
  if (countDiffPixels) {
    defChangedPixelsLum = 1.0 / (long double)countDiffPixels * sum2Lum -
      sumLum * sumLum / (long double)countDiffPixels /
      (long double)countDiffPixels;
    if (defChangedPixelsLum > 0.f)
      defChangedPixelsLum = sqrt(defChangedPixelsLum);
    else
      defChangedPixelsLum = 0;
    defChangedPixelsRGB = 1.0 / (long double)countDiffPixels * sum2RGB -
      sumRGB * sumRGB / (9*(long double)countDiffPixels *
			 (long double)countDiffPixels);
    if (defChangedPixelsRGB > 0.f)
      defChangedPixelsRGB = sqrt(defChangedPixelsRGB);
    else
      defChangedPixelsRGB = 0;
    
    RMSErrorChangedPixelsLum = sqrt(sum2Lum / (long double)(countDiffPixels));
    RMSErrorChangedPixelsRGB = sqrt(sum2RGB / (long double)(countDiffPixels*3));
  }

  // computing PSNR - peak signal-to-reconstruced image measure
  long double PSNR = 20.f * log10(255.0 / RMSErrorLum);
  long double PSNR_RGB = 20.f * log10(255.0 / RMSErrorRGB);
  if (verbose) {
    cout << "Comparing two images by RMSE (root-mean-squared) error"
	 << endl;
    cout << "Percentage of changed pixels is "
	 << 100.0f * countDiffPixels / countPixels << " [%] ("
	 << countDiffPixels << " from cntAllPixels = " << countPixels
	 << ")" << endl;
    cout << "For all pixels RMSE error[RGB,LUM] = "
	 << RMSErrorRGB * 100.0f << " " << RMSErrorLum * 100.0f << " [%] " << endl; 
    cout << "For only changed pixels RMSE error = "
	 << RMSErrorChangedPixelsRGB * 100.0f << " " << RMSErrorChangedPixelsLum * 100.0f << " [%] " << endl; 
    cout << "Standard deviation across whole image is = " << devRGB << " " << devLum << endl;
    cout << "Standard deviation for changed pixels only image is = "
	 << defChangedPixelsRGB << " " << defChangedPixelsLum << endl;
    cout << "For all pixels PSNR = " << PSNR_RGB << " " << PSNR << " [decibels] " << endl;
    cout << "Main result from comparing two images = " << RMSErrorLum << endl;
  }
  return differ; // the same size .. false, different size .. true
}

// ------------------------------------------------------------------------------------

void
CEnvironmentMap::setFunctionConstNorthHemisphere(float r, float g, float b)
{
  // Copy the data from PNG representation
  for (int y=0; y < height; y++) {
    float theta = yToTheta(y);
    float cosTheta = fabs(cos(theta))+1e-6;
    if (theta <= M_PI/2.f) {
      assert(cosTheta >= 0.f);
      for (int x=0; x < width; x++) {
	int ww = (width * y + x)*3;
	// set it to the output format as required by RGBE_WRITE
	cols[ww+0] = r;
	cols[ww+1] = g;
	cols[ww+2] = b;
      } // for jcol
    } // if north hemisphere
    else {
      // South hemisphere set to almost zero value
      const float almostZero = 1e-20;
      for (int x=0; x < width; x++) {
	int ww = (width * y + x)*3;
	// set it to the output format as required by RGBE_WRITE
	cols[ww+0] = almostZero*cosTheta;
	cols[ww+1] = almostZero*cosTheta;
	cols[ww+2] = almostZero*cosTheta;
      } // for x
    } // if south hemisphere
  } // for y
}

void
CEnvironmentMap::setFunctionConstSouthHemisphere(float r, float g, float b)
{
  // Copy the data from PNG representation
  for (int y=0; y < height; y++) {
    float theta = yToTheta(y);
    float cosTheta = fabs(cos(theta))+1e-6;
    if (theta > M_PI/2.f) {
      // if north hemisphere
      const float almostZero = 1e-20;
      assert(cosTheta >= 0.f);
      for (int x=0; x < width; x++) {
	int ww = (width * y + x)*3;
	// set it to the output format as required by RGBE_WRITE
	cols[ww+0] = r;
	cols[ww+1] = g;
	cols[ww+2] = b;
      } // for jcol
    } 
    else {
      // if north hemisphere - do nothing
    }
  } // for y
}

void
CHDRImage::copyDataFromArray(int nwidth, int nheight, int channels,
			     float *buffer, int bufferwidth,
			     int xoffset, int yoffset,
			     bool flipVertically)
{
  // Copy the data from buffer
  for (int y=0; y < nheight; y++) {
    if (y+yoffset < 0)
      continue;
    if (y+yoffset >= height)
      break;
    for (int x=0; x < nwidth; x++) {
      if (x+xoffset < 0)
	continue;
      if (x+xoffset >= width)
	break;
      // index for storing data
      int si = (width * (y+yoffset) + x+xoffset);
      // index for reading data
      int ri;
      if (flipVertically) // flipping image vertically
	ri = (bufferwidth * (nheight - y) + x);
	  else
        ri = (bufferwidth * y + x);
      if ((channels == 3)||(channels==4)) {
	//indexes are the same, it can be either RGB or RGBA image
	// our image here has 3 channels
	// source buffer has 3 channels
	si *= 3;
	ri *= channels;
	cols[si] = buffer[ri];
	cols[si+1] = buffer[ri+1];
	cols[si+2] = buffer[ri+2];
      } else {
	// read index different to write index
	// our image here has 3 channels
	// source buffer has 1 channels
	si *= 3;
	ri *= channels;
	cols[si] = buffer[ri]; // copy the first channel to all three channels
	bool copyToAll = true;
	if (copyToAll) {
	  cols[si+1] = buffer[ri];
	  cols[si+2] = buffer[ri];
	}
	else {
	  if (channels == 2) {
	    // we have also the second channel
	    cols[si+1] = buffer[ri+1];	  
	  }
	  cols[si+2] = 0; // the last channel set to zero
	}
      } // channel = 1 or channel = 2
    } // for jcol
  } // for irow  
}


// -------------------------------------------------------------
// Environment map

CEnvironmentMap::~CEnvironmentMap()
{
  delete []lums; lums = 0;
  delete []cumlums; cumlums = 0;
#ifdef FIRST_ROWS
  // summing up rows vertically in the second pass
  delete []cumrows; cumrows = 0;
#else // FIRST_ROWS
  // summing up columns in the horizontal direction in the second pass
  delete []cumcols; cumcols = 0;
#endif // FIRST_ROWS
  delete []meanAndVarPartialDirVal;
  meanAndVarPartialDirVal = 0;
  deallocatePrecomputedLights();
}

CEnvironmentMap::CEnvironmentMap():
  CHDRImage(), lums(0), cumlums(0),
#ifdef FIRST_ROWS
  cumrows(0),
#else
  cumcols(0),
#endif
  totalPower(0),
  totalPowerOrig(-100),
  envLightPositions(0),
  envLightColours(0),
  envLightPositionsBackup(0),
  envLightColoursBackup(0),
  envLightProbs(0),
  cntAllocatedLights(0),
  cntLights(0),
  varianceWholeSphere(-1.f),
  cntDirsVDmax(0),
  meanAndVarPartialDirVal(0),
  randomizePrecomputedLights(false)
{
}

void
CEnvironmentMap::deallocatePrecomputedLights()
{
  // data for the lights
  freemem2(envLightPositions, 0, cntAllocatedLights-1, 0, 2);
  envLightPositions = 0;
  freemem2(envLightPositionsBackup, 0, cntAllocatedLights-1, 0, 2);
  envLightPositionsBackup = 0;
  freemem2(envLightColours, 0, cntAllocatedLights-1, 0,2);
  envLightColours = 0;
  freemem2(envLightColoursBackup, 0, cntAllocatedLights-1, 0,2);
  envLightColoursBackup = 0;
  freemem1(envLightProbs, 0, cntAllocatedLights-1);
  envLightColoursBackup = 0;
  cntAllocatedLights = 0;
  cntLights = 0;
}

// deep copy constructor
CEnvironmentMap::CEnvironmentMap(const CEnvironmentMap &src):
  CHDRImage(src)
{
  computeLuminance();
  computeCDF();
  cntLights = src.cntAllocatedLights;
  if (cntLights) {
    allocatePrecomputedLights(src.cntAllocatedLights);  
    // Precomputed lights on the fly and store them to the array
    // float sumProb = 0.f;
    // Let us normalize to the unit power from the whole sphere
    for(int i=0; i < cntLights; i++) {
      // Positions
      envLightPositions[i][0] = src.envLightPositions[i][0];
      envLightPositions[i][1] = src.envLightPositions[i][1];
      envLightPositions[i][2] = src.envLightPositions[i][2];
      // Positions backups
      envLightPositionsBackup[i][0] = src.envLightPositionsBackup[i][0];
      envLightPositionsBackup[i][1] = src.envLightPositionsBackup[i][1];
      envLightPositionsBackup[i][2] = src.envLightPositionsBackup[i][2];
      // Colors
      envLightColours[i][0] = src.envLightColours[i][0];
      envLightColours[i][1] = src.envLightColours[i][1];
      envLightColours[i][2] = src.envLightColours[i][2];
      // Probabality density for the light
      envLightProbs[i] = src.envLightProbs[i];
      // Make a backup of colours
      envLightColoursBackup[i][0] = src.envLightColoursBackup[i][0];
      envLightColoursBackup[i][1] = src.envLightColoursBackup[i][1];
      envLightColoursBackup[i][2] = src.envLightColoursBackup[i][2];
    } // no of lights
  }
}

void
CEnvironmentMap::allocatePrecomputedLights(int cnt)
{
  cntAllocatedLights = cnt;
  envLightPositions = allocation2(0, cntAllocatedLights-1, 0, 2);
  assert(envLightPositions);
  envLightPositionsBackup = allocation2(0, cntAllocatedLights-1, 0, 2);
  assert(envLightPositionsBackup);
  envLightColours = allocation2(0, cntAllocatedLights-1, 0, 2);
  assert(envLightColours);
  envLightColoursBackup = allocation2(0, cntAllocatedLights-1, 0, 2);
  assert(envLightColoursBackup);
  envLightProbs = allocation1(0, cntAllocatedLights-1);
  assert(envLightProbs);
}  

// Compute luminance from color values
void
CEnvironmentMap::computeLuminance()
{
  delete []lums;
  lums = new float[height * width];
  assert(lums);

  const int maxindex = 3 * width * height;
  for (int h = 0; h < height; h++) {
    for (int w = 0; w < width; w++) {
      int wl = h*width+w;
      int wc = wl*3;
      assert(wc+2 < maxindex);
      float lum = luminance(&(cols[wc]));
      lums[wl] = lum; // set the luminance
    } // w
  } // h

  return;
}

// Normalize the power of EM map so the integral is 1 Watt.
void
CEnvironmentMap::normalizePower()
{
  if (fabs(totalPower - 1.0f) < 1e-2) {
    // the power of EM was already normalized or is accidentally 1.0
    return;
  }
  // if luminance was not computed, do it now
  if (lums == 0)
    computeLuminance();
  int maxIterations = 20;
  for (int i = 0; i < maxIterations; i++) {
    computeTotalPower();
    // cout << "i= " << i << " ... totalPower before normalization = " << totalPower << endl;
    assert(totalPower > 0.f);
    float mult = 1.0f/totalPower;
    // normalize the values so the totalPower = 1.0
    const int maxIndex = width*height*3;
    for (int j = 0; j < maxIndex; j++)
      cols[j] *= mult;
    // compute the luminance and reevaluate the power
    computeLuminance();
    computeTotalPower();
    if (fabs(totalPower - 1.0f) < 1e-2) {
      cout << "Environment map was successfully normalized in " << i+1
	   << " iterations to power = " << totalPower << endl;
      return;
    }
    else {
      cout << "Current result in the implementation of normalization of total power " << endl;
      cout << "totalPower = " << totalPower << endl;
    }
  } // for i
  cout << "Environment map was normalized in " << maxIterations
       << " iterations to power = " << totalPower << endl;
  return;
}


void
CEnvironmentMap::computeTotalPower()
{
#define DETERMINISTIC
#ifdef DETERMINISTIC
  // There is a bug in this code, when integrating 1 over the sphere,
  // we are not getting 4*PI, Vlastimil Havran, 2013-09-10
  
  // Fast version with precomputation of sin and cos arrays
  int iphi, itheta;
  double a1, a2, b1, b2, dp, dt, aa, bb;

#ifdef _COMPUTE_MAXINTENSITY
  // the information to be computed is initialized
  float _maximum_d_INTENSITY = -1.0;
#endif  
  this->totalPower = 0.0;

  // Precompute sinus and cosinus for theta
  float *sinDT = new float[height];
  assert(sinDT);
  long double sum = 0.;
  // precompute sine and cosine of theta
  for (itheta = 0; itheta < height ; ++itheta)
  {
    float theta = (float)M_PI*(itheta+0.5f)/height;
    sinDT[itheta] = sin(theta);
    //cosDT[itheta] = cos(theta);
  } // for itheta
  
  // for all patches on the goniometric web
  for (iphi = 0; iphi < width; ++iphi) {
    for (itheta = 0; itheta < height - 1; ++itheta) {
      // the four values for given goniometric web diagram patch
      int iphiRight = (iphi + 1) % width;
      // The densities for the four corners of the goniometric patch
      a1 = lums[addrlum(iphi, itheta)];
      a2 = lums[addrlum(iphiRight, itheta)];
      b1 = lums[addrlum(iphi, itheta)];
      b2 = lums[addrlum(iphiRight, itheta+1)];
#ifdef _COMPUTE_MAXINTENSITY
      // Find out the maximum for statistic purposes
      dt = a1;
      if (a2 > dt)
	dt = a2;
      if (b1 > dt)
	dt = b1;
      if (b2 > dt)
	dt = b2;
      if (dt > _maximum_d_INTENSITY)
        _maximum_d_INTENSITY = dt;
#endif      
      // delta theta
      dt = M_PI/(float)height;
      // deltha phi
      dp = 2.f*M_PI/(float)width;
      // for a single patch inaccurate computation - we average the intensities
      // and multiply the result by the area of the patch dt*dp and also by sin(theta)
      double deltaPower = sinDT[itheta]*(a1+a2+b1+b2)*0.25*dt*dp;
      assert(deltaPower >= 0.f);
      sum += deltaPower;
    } // for itheta
  } // for iphi
  totalPower = sum;
  delete []sinDT;
#else
  // RANDOM evaluation
  int cnt = 1000000;
  long double sum = 0.;
  float lum;
  for(int i = 0; i < cnt; i++) {
    //Generate random direction on unit sphere with density prob(theta) = 1/(2*PI).
    CVector3D result;
    //CVector2D rndvalue;
    //rndvalue.xx = RandomValue();
    //rndvalue.yy = RandomValue();
    // constant density on surface of sphere
    result = RND::sph_const(); 
    //result = RND::sph_const(rndvalue);
    float XYZ[3];
    XYZ[0] = result.x;
    XYZ[1] = result.y;
    XYZ[2] = result.z;
    float theta, phi;
    ConvertXYZtoThetaPhi2(XYZ, theta, phi); // in radians
    float RGB[3];
    this->getRGB_ThetaPhi(theta, phi, RGB);
    float lum = luminance(RGB);
    //    cout << "lum = " << lum << " ";
    if ((lum < 0.f)||(isnan_local(lum))) {
      cout << "sum = " << sum << endl;
      cout << "Problem = " << lum << endl;
      exit(0);
    }
    sum += (long double)lum;
    //cout << sum << " ";
  }
  //  cout << "sum = " << sum << " cnt" << cnt << endl;
  sum /= (long double) cnt;
  const float SAHEM = 4.0f*M_PI; // the surface area of the sphere
  totalPower = SAHEM * sum;
#endif

  if (fabs(totalPowerOrig+100.f) < 1e-5) {
    // save original total power prior to normalization only once
    totalPowerOrig = totalPower;
  }
  
  totalPowerInv = 1.0f/totalPower;

  cout << "TotalPower = " << totalPower << " Watts" << endl;
  return;
}

// Compute CDF from the luminance, luminance has to be set
bool
CEnvironmentMap::computeCDF()
{
  std::cout << "Starting to compute CUM image " << std::endl;
  cumlums = new double[height * width];
  assert(cumlums);

#ifdef FIRST_ROWS
  assert(cumrows == 0);
  cumrows = new double[height+2];
  assert(cumrows);
  // The first rows are summed up in horizontal direction,
  // each individually. This is the first row.
  float mult = 0.f;
  mult = sin(0.5*M_PI/((float)height))/float(height);
  cumlums[0] = lums[0] * mult;
  for (int w = 1; w < width; w++) {
    //std::cout << lums[w] << " ";
    cumlums[w] = cumlums[w-1] + lums[w] * mult;
  } // w
  // the first value to the vertical direction
  cumrows[0] = cumlums[width-1];
  assert(cumrows[0] >= 0.f);

  // all other rows of an environment image
  for (int h = 1; h < height; h++) {
    int windex = addrcum(0, h);
    mult = sin((0.5f+(float)h)*M_PI/((float)height));
    cumlums[windex] = mult*lums[windex];
    windex++;
    // a single row summing up
    for (int w = 1; w < width; w++) {
      //std::cout << lums[width-1] << " ";
      // this is conditional probability for the first dimension
      cumlums[windex] = cumlums[windex-1] + mult * lums[windex];
      assert(addrcum(w, h) == windex);
      windex++;
    } // w
    //std::cout << cumlums[windex-1] << " ";
    // This corresponds to joint probability for the second dimension
    cumrows[h] = cumrows[h-1] + cumlums[windex-1];
    assert(cumrows[h] >= 0.f);
    //std::cout << cumrows[h] << " ";
  } // h
  cout << "cumrows[height-1]= " << cumrows[height-1] << endl;  
#else
  // as in PBRT book
  assert(cumcols == 0);
  cumcols = new double[width+2];
  assert(cumcols);
  // The first rows are summed up in horizontal direction,
  // each individually, this is the first row
  float mult = 0.f;
  mult = sin(0.5*M_PI/((float)height));
  cumcols[0] = lums[0] * mult;
  for (int h = 1; h < height; h++) {
    //std::cout << lums[w] << " ";
    int hi = addrcum(0, h);
    int hi2 = addrcum(0, h-1);
    int li = addrlum(0, h);
    cumlums[hi] = cumlums[hi2] + lums[li] * mult;
  } // w
  // the first value to the vertical direction
  int hi = addrlum(0, height-1);
  cumcols[0] = cumlums[hi];
  assert(cumcols[0] >= 0.f);

  // all other rows of an environment image
  for (int w = 1; w < width; w++) {
    int wi = addrcum(w, 0);
    int li = addrlum(w, 0);
    mult = sin(0.5f*M_PI/((float)height));
    cumlums[wi] = lums[li] * mult;
    // a single row summing up
    for (int h = 1; h < height; h++) {
      mult = sin((0.5f+(float)h)*M_PI/(float)height);
      //std::cout << lums[width-1] << " ";
      // this is conditional probability for the first dimension
      wi = addrcum(w, h);
      int wi2 = addrcum(w, h-1);
      li = addrlum(w, h);
      cumlums[wi] = cumlums[wi2] + mult * lums[li];
    } // w
    //std::cout << cumlums[windex-1] << " ";
    // This corresponds to joint probability for the second dimension
    wi = addrcum(w, height-1);
    cumcols[w] = cumcols[w-1] + cumlums[wi];
    assert(cumcols[w] >= 0.f);
    //std::cout << cumrows[h] << " ";
  } // h
  cout << "cumcols[width-1]= " << cumcols[width-1] << endl;
#endif  

  // std::cout << "Ending to compute CUM image " << std::endl;

  // finally, compute total power from EM
  computeTotalPower();
  return false;
}


bool
SingleBinSearch(double list[], int n, float target, // input
		int &nindex, float &alpha) // output
{
  //std::cout << list[0] << " " << target << " " << list[n-1] << std::endl;
  if (target < list[0]) {
    //cout << "SBS1a" << endl;
    nindex = 0;
    alpha = target/(list[0]+1e-6);
    return false; // finished
  }
  if (target > list[n-1]) {
    cout << "SBS1b" << endl;
    return true; // cannot be found
  }
  //  cout << "BB" << endl;
  //declare low, mid, and high
  int low = 0, mid, high = n-1;
#if 1
  for (int i = low; i < high; i++) {
    if (list[i] > list[i+1]) {
      cout << "A The array for binary search is not sorted" << endl;
      abort();
    }
  }
#endif  
  //binary search loop
  while(low <= high) {
    //compute mid
    mid = (low + high) / 2;
    //cout << "low = " << low << " " << list[low] << " high= " << high << " " << list[high] << endl;
    //cout << "mid = " << mid << " " << list[mid] << " target = " << target << endl;
    //check target against list[mid]
    if (target < list[mid]) {
      //if target is below list[mid] reset high
      high = mid-1;
      //cout << "AAAA" << endl;
    }
    else if (target > list[mid]) {
      //if target is below list[mid]) reset low
      low = mid+1;
      //cout << "BBB" << endl;
    }
    else {
      //if target is found set low to exit the loop
      low = mid;
      high = mid-1;
      //cout << "CCC" << endl;
    }
  } // while
  //cout << "CC" << endl;
  // This condition below is necessary
  if (list[mid] > target)
    mid--;
  
  assert(0 <= mid);
  assert(mid < n);
  //cout << "list[mid] " << list[mid]
  //     << " target = " << target
  //     << " list[mid+1] " << list[mid+1] << endl;
  assert(list[mid] <= target);
  assert(target <= list[mid+1]);
  if (mid < n-1) {
    nindex = mid;
    alpha = (target - list[mid])/(list[mid+1] - list[mid]);
    assert(alpha >= 0);
    assert(alpha <= 1.f);
    return false; // OK
  }

  //cout << "DDD" << endl;
  nindex = n-2;
  alpha = 1;
  return false; // OK
}

bool
DoubleBinSearch(double list1[], double list2[], // input two arrays with float values
  float alpha, int n, float target, // input alpha for mixing, number of values, search value
  int &nindex, float &beta) // output
{
  float valueMin = list1[0]*(1.f-alpha)+list2[0]*alpha;
  if (target < valueMin) {
    //cout << "DB2a" << endl;
    nindex = 0;
    alpha = target/(valueMin+1e-6);
    return false; // OK
  }
  float valueMax = list1[n-1]*(1.f-alpha)+list2[n-1]*alpha;
  if (valueMax < target) {
    cout << "DB2b" << endl;
    cout << "target = " << target << " valMax = " << valueMax << endl;
    return true; // cannot be found
  }

  //declare low, mid, and high
  int low = 0, mid, high = n-1;

#if 1
  for (int i = low; i < high; i++) {
    if (list1[i] > list1[i+1]) {
      cout << "BThe array for binary search is not sorted" << endl;
      abort();
    }
    if (list2[i] > list2[i+1]) {
      cout << "CThe array for binary search is not sorted" << endl;
      abort();
    }
  }
#endif  
  
  //binary search loop
  while (low <= high) {
    //compute mid
    mid = (low + high) / 2;
    //check target against list[mid]
    float value = list1[mid]*(1.f-alpha)+list2[mid]*alpha;
    if (target < value)
      //if target is below list[mid] reset high
      high = mid-1;
    else if (target > value)
      //if target is below list[mid]) reset low
      low = mid+1;
    else {
      //if target is found set low to exit the loop
      low = mid;
      high = mid-1;
    }
  } // while
  // This condition below is necessary
  if (list1[mid]*(1.f-alpha)+list2[mid]*alpha > target)
    mid--;
  
  float value = list1[mid]*(1.f-alpha)+list2[mid]*alpha;
  float value2 = list1[mid+1]*(1.f-alpha)+list2[mid+1]*alpha;
  //  cout << "lowval = " << value << " target = " << target << " hival = " << value2 << endl;
  assert(0 <= mid);
  assert(mid < n);
  assert(value <= target);

  if (mid < n-1) {
    nindex = mid;
    assert(target <= value2);
    beta = (target - value)/(value2-value+1e-30);
    assert(beta >= 0.f);
    assert(beta <= 1.f);
    return false; // OK
  }
  
  nindex = n-2;
  beta = 1.f;
  return false; // OK
}

// get value RGB for a particular theta and phi, when we assume
// that the HDR image is the environment map mapped to a sphere
bool
CEnvironmentMap::importanceSamplingXY(float rnd1, float rnd2, // input
				      float &X, float &Y, // output position in the image in pixels
				      float RGB[3]) const
{
  assert(0 <= rnd1); assert(rnd1 <= 1);
  assert(0 <= rnd2); assert(rnd2 <= 1);

#ifdef FIRST_ROWS  
  // first we search along theta
  // theta = 0 .. north pole
  // theta = PI .. south pole
  float valueSearch = rnd1 * cumrows[height-1];
  //std::cout << "cumrows[height-1] = " << cumrows[height-1] << std::endl;
  int y = 0; // index in range 0 ... height-2
  float betaY = 0.f; // value in range 0 ... 1
  bool res1 =
    SingleBinSearch(cumrows, height, valueSearch, y, betaY);
  assert(res1 == false);
  Y = (float)(y+betaY);
  assert(0 <= y);
  assert(y < height);
  int y2 = (y+1);
  if (y2 >= height) y2 = height;
  float valueSearch2 = rnd2 * (cumlums[y*width+width-1]*(1.f-betaY) + 
			       cumlums[y2*width+width-1]*betaY);
  int x = 0; // index in range 0 ... width - 2
  float alphaX = 0.f; // value in range 0 ... 1
  bool res2 =
    DoubleBinSearch(&(cumlums[y*width]), 
		    &(cumlums[(y+1)*width]),
		    betaY, width, valueSearch2, x, alphaX);
  assert(res2 == false);
  assert(0 <= x);
  assert(x < width-1);
  X = (float)x + alphaX;
  assert(X >= 0.f);
  assert(X < width);
  int x2 = (x+1) % width;
#else // FIRST_ROWS  
  // First search to determine X
  // first we search along phi
  // theta = 0 .. north pole
  // theta = PI .. south pole
  float valueSearch = rnd1 * cumcols[width-1];
  //std::cout << "cumcols[height-1] = " << cumcols[height-1] << std::endl;
  int x = 0; // index in range 0 ... height-2
  float alphaX = 0.f; // value in range 0 ... 1
  bool res1 =
    SingleBinSearch(cumcols, width, valueSearch, x, alphaX);
  assert(res1 == false);
  X = (float)(x+alphaX);
  int x2 = (x+1)%width;
  assert(0 <= x);
  assert(x < width);
  // Then search in two particular columns to determine Y
  float valueSearch2 = rnd2 * (cumlums[addrcum(x, height-1)]*(1.f-alphaX) + 
			       cumlums[addrcum(x2, height-1)]*alphaX);
  int y = 0; // index in range 0 ... height-2
  float betaY = 0.f; // value in range 0 ... 1

  bool res2 =
    DoubleBinSearch(&(cumlums[addrcum(x, 0)]), &(cumlums[addrcum(x2,0)]),
		    alphaX, height, valueSearch2, y, betaY);
  assert(res2 == false);
  assert(0 <= y);
  assert(y < height-1);
  Y = (float)y + betaY;
  assert(Y >= 0.f);
  assert(Y < height);
  int y2 = (y+1);
  if (y2 >= height) y2 = height;
#endif
  // The colour is taken from the particular pixel
  // by interpolation from the 4 neighboring pixels
  
  for (int i = 0; i < 3; i++) {
    RGB[i] = cols[(width*y + x)*3 +i]*(1.f-betaY)*(1.f-alphaX)
      + cols[(width*y + x2)*3 + i]*(1.f-betaY)*alphaX
      + cols[(width*y2 + x)*3 + i]*betaY*(1.f - alphaX)
      + cols[(width*y2 + x2)*3 + i]*betaY*alphaX;
    //cout << RGB[i] << " ";
  } // for i

  float lum = luminance(RGB)+1e-20;
  assert(lum > 0.f);
  float norm = 1.0f/lum * totalPower;
  for (int i = 0; i < 3; i++) {
    RGB[i] *= norm;
  }

  // now the color is of unit importance(=luminance)
  return false;
}

bool
CEnvironmentMap::saveDebugHDR(char *fileName, int mode, float mult)
{
  CHDRImage img;
  img.allocate(width, height);

  if (mode == 0) {
    assert(cols);
    int nmax = 3*width*height;
    // saving color image
    int ii=0;
    for (int h = 0; h < height; h++) {
      ii = width * h * 3;
      for (int w = 0; w < width; w++) {
	float red = mult * cols[ii]; ii++;
	float green = mult * cols[ii]; ii++;
	assert(ii  < nmax);
	float blue = mult * cols[ii]; ii++;
	img.setRGB_XY(w, h, red, green, blue);
      } // w
    } // h
    img.saveHDR(fileName);
    return false;
  }
  if (mode == 1) {
    assert(lums);
    // saving luminance image
    int ii=0;
    for (int h = 0; h < height; h++) {
      for (int w = 0; w < width; w++) {
	float lum = mult * lums[ii++]; 
	img.setRGB_XY(w, h, lum, lum, lum);
      } // w
    } // h
    img.saveHDR(fileName);
    return false;
  }
  if (mode == 2) {
    assert(cumlums);
    // saving cum sum of luminance image
    int ii=0;
    for (int h = 0; h < height; h++) {
      for (int w = 0; w < width; w++) {
	float val = mult * cumlums[ii++]; 
	img.setRGB_XY(w, h, val, val, val);
      } // w
    } // h
    img.saveHDR(fileName);
    return false;
  }
  std::cout << "ERROR: unknown mode for saving image" << std::endl;
  return true;
}

// get value RGB for a particular theta and phi, when we assume
// that the HDR image is the environment map mapped to a sphere
bool
CEnvironmentMap::importanceSamplingThetaPhi(float rnd1, float rnd2, // input
					    float &theta, float &phi, // output
					    float RGB[3], float &prob) const
{
  float X = 0.f, Y = 0.f;
  bool res = importanceSamplingXY(rnd1, rnd2, X, Y, RGB);
  assert(res == false);
  // now convert X/Y of the image to theta/phi
  phi = xToPhi(X); // 0 to 2*PI
  assert(0.f <= phi);
  assert(phi <= 2.f*M_PI);
  theta = yToTheta(Y); // 0 to PI

  prob = computeProb(X, Y, theta, RGB);
  return false; // OK
}

// precompute cnt lights
void
CEnvironmentMap::precomputeLights(int cnt)
{
  if (cnt > cntAllocatedLights) {
    deallocatePrecomputedLights();
    allocatePrecomputedLights(cnt);
  }
  cntLights = cnt;
  
  // Precomputed lights on the fly and store them to the array
  float sumProb = 0.f;
  static float RGB[3], XYZ[3];
  // Let us normalize to the unit power from the whole sphere
  for(int i=0; i < cntLights; i++) {
    float rnd1 = RandomValue();
    float rnd2 = RandomValue();
    float theta, phi, X, Y, prob;
    importanceSamplingThetaPhi(rnd1, rnd2, theta, phi, RGB, prob);
    ConvertThetaPhiToXYZ(theta, phi, XYZ);
    //environMap[em].importanceSamplingXY(rnd1, rnd2, X, Y, RGB);	  
    //environMap[em].MakeMark(X, Y, EE_BRDFSamples, 10, 100.f);    
    // Positions
#ifdef XYZCOORSYSTEM
    // Z-axis is in the direction of north-pole
    // plane XY corresponds to theta=PI/2 and phi<0,2PI)
    envLightPositionsBackup[i][0] = envLightPositions[i][0] = XYZ[0];
    envLightPositionsBackup[i][1] = envLightPositions[i][1] = XYZ[1];
    envLightPositionsBackup[i][2] = envLightPositions[i][2] = XYZ[2];
#else
    // Y-axis is in the direction of north-pole
    // plane XZ corresponds to theta=PI/2 and phi<0,2PI)
    envLightPositionsBackup[i][0] = envLightPositions[i][0] = XYZ[0]; // X axis remains
    envLightPositionsBackup[i][1] = envLightPositions[i][1] = XYZ[2]; // Z axis become Y
    envLightPositionsBackup[i][2] = envLightPositions[i][2] = XYZ[1]; // Y axis becomes X
#endif    
    // Colors
    envLightColours[i][0] = RGB[0];
    envLightColours[i][1] = RGB[1];
    envLightColours[i][2] = RGB[2];
    // Probabality density for the light
    envLightProbs[i] = prob;
#if 0
    cout << "i = " << i << " Theta=" << theta/M_PI*180.f << " Phi= " << phi/M_PI*180.f << " "
	 << " RGB= " << envLightColours[i][0] << " " << envLightColours[i][1]
	 << " " << envLightColours[i][2] << " prob=" << prob << endl;
#endif	
    // Make a backup of colours
    envLightColoursBackup[i][0] = envLightColours[i][0];
    envLightColoursBackup[i][1] = envLightColours[i][1];
    envLightColoursBackup[i][2] = envLightColours[i][2];
    
    sumProb += prob; // adding the probabilities
  } // no of lights
  cout << "sumProb = " << sumProb/float(cntLights) << endl;

  //exit(0);
  printf("The directional lights were computed now\n");

  return;
}

void
CEnvironmentMap::makeLighter(int i, float c1, float c2)
{
  float lum = luminance(envLightColours[i]);
  for (int j = 0; j < 3; j++)
    envLightColours[i][j] = lum * c1 + envLightColoursBackup[i][j] * c2;
}

// It rotates only the directional lights, not the image!
void 
CEnvironmentMap::rotateEnvironmentLights(float angle)
{
  cout << "Rotate EL angle = " << angle << endl;
  
  float sv[2];
  // ... rotation of whole environment lights by angle
  float sinAngle = sin(angle);
  float cosAngle = cos(angle);
  cout << "cosAngle = " << cosAngle << " sinAngle = " << sinAngle << endl;
  // here we access direction directly to arrays, no randomization applies
  for(int i=0; i < cntLights; i++) {
#ifdef XYZCOORSYSTEM
    // Z-axis is in the direction of north-pole
    // plane XY corresponds to theta=PI/2 and phi<0,2PI)
    const int Xi=0, Yi = 1;
#else
    // Y-axis is in the direction of north-pole
    // plane XZ corresponds to theta=PI/2 and phi<0,2PI)
    const int Xi=0, Yi = 2;
#endif
    sv[0] = cosAngle * envLightPositionsBackup[i][Xi] - sin(angle) * envLightPositionsBackup[i][Yi];
    sv[1] = sinAngle * envLightPositionsBackup[i][Xi] + cos(angle) * envLightPositionsBackup[i][Yi];
    envLightPositions[i][Xi] = sv[0];
    envLightPositions[i][Yi] = sv[1];
  } // for i
#if 0
  // disable randomization for printing
  bool save = randomizePrecomputedLights;
  randomizePrecomputedLights = false;
  for(int i=0;i<cntLights;i++) {
    float x = getDirectionPrecomputedLight(i, 0);
    float y = getDirectionPrecomputedLight(i, 1);
    float z = getDirectionPrecomputedLight(i, 2);
    //      cout << "x = " << x << "y = " << y << " z = " << z << endl;
    //      cout << "phi = " << phi << " theta = " << theta << endl;
    //      cout << " lum = " << lum << " prob = " << prob << " lum/prob = " << lum/prob << endl;
  } // for i
  randomizePrecomputedLights = save;
#endif  
  return;
}//--- rotateEnvironmentLights ---------------------------------------------

// get the color when indexed as image
inline bool
CEnvironmentMap::setLUM_XY(int x, int y, float val)
{
  if ((x < 0)||(x >= width)) return true; // error
  if ((y < 0)||(y >= height)) return true; // error
  int itt = (width*y+x);
  lums[itt] = val;
  return false;
}

// make mark in luminance channel
void
CEnvironmentMap::makeMarkLUM(int mx, int my, EMarkType typeMark,
			     int marksize, float mult)
{
  int x, y;
  float val;
  if (typeMark == EE_BRDFSamples) {
    val = 255.0; // white
  }
  else {
    // ENVMAP samples
    val = 255.0;
  }
  // Make mark with correct color
  val *= mult;
  if (typeMark == EE_BRDFSamples) {
    // the horizontal direction
    if ((my >= 0) && (my < height)) {
      for (x = mx - marksize; x <= mx + marksize; x++) {
	if ((x < 0) || (x >= width))
	  continue;
	setLUM_XY(x, my, val);
      }
    }
    if ((mx >= 0) && (mx < width)) {
      for (y = my - marksize; y <= my + marksize; y++) {
	if ((y < 0) || (y >= height))
	  continue;
	setLUM_XY(mx, y, val);
      }
    }
  }
  else {
    int t;
    // from left bottom to right top corner
    for (t = - marksize; t <= marksize; t++) {
      x = mx + t;
      if ((x < 0) || (x >= width))
	continue;
      y = my + t;
      if ((y < 0) || (y >= height))
	continue;
      setLUM_XY(x, y, val);
    } // for t
    // from left top to right bottom corner
    for (t = - marksize; t <= marksize; t++) {
      x = mx + t;
      if ((x < 0) || (x >= width))
	continue;
      y = my - t;
      if ((y < 0) || (y >= height))
	continue;
      setLUM_XY(x, y, val);
    } // for t
  }

  return; // mark finished
}

void
CEnvironmentMap::setFunctionCosineTheta(float r, float g, float b)
// for the whole sphere!
{
  // Copy the data from PNG representation
  for (int y=0; y < height; y++) {
    float theta = yToTheta(y);
    float cosTheta = fabs(cos(theta)); // absolute value
    for (int x=0; x < width; x++) {
      int ww = (width * y + x)*3;
      // set it to the output format as required by RGBE_WRITE
      cols[ww+0] = r*cosTheta;
      cols[ww+1] = g*cosTheta;
      cols[ww+2] = b*cosTheta;
    } // for jcol
  } // for irow
  return;
}

void
CEnvironmentMap::setFunctionCosineThetaNorthHemisphere(float r, float g, float b)
{
  // Copy the data from PNG representation
  for (int y=0; y < height; y++) {
    float theta = yToTheta(y);
    float cosTheta = fabs(cos(theta))+1e-6;
    if (theta <= M_PI/2.f) {
      assert(cosTheta >= 0.f);
      for (int x=0; x < width; x++) {
	int ww = (width * y + x)*3;
	// set it to the output format as required by RGBE_WRITE
	cols[ww+0] = r*cosTheta;
	cols[ww+1] = g*cosTheta;
	cols[ww+2] = b*cosTheta;
      } // for jcol
    } // if north hemisphere
    else {
      // South hemisphere set to almost zero value
      const float almostZero = 1e-20;
      for (int x=0; x < width; x++) {
	int ww = (width * y + x)*3;
	// set it to the output format as required by RGBE_WRITE
	cols[ww+0] = almostZero*cosTheta;
	cols[ww+1] = almostZero*cosTheta;
	cols[ww+2] = almostZero*cosTheta;
      } // for x
    } // if south hemisphere
  } // for y
  return;
}

void
CEnvironmentMap::setFunction2(float r, float g, float b)
{
  // Copy the data from PNG representation
  for (int y=0; y < height; y++) {
    float theta = yToTheta(y);
    float value = fabs(sin(theta))+1e-6;
    for (int x=0; x < width; x++) {
      int ww = (width * y + x)*3;
      // set it to the output format as required by RGBE_WRITE
      cols[ww+0] = r*value;
      cols[ww+1] = g*value;
      cols[ww+2] = b*value;
    } // for jcol
  } // for irow
}

void
CEnvironmentMap::setFunction3(float r, float g, float b)
{
  // Copy the data from PNG representation
  for (int y=0; y < height; y++) {
    float theta = yToTheta(y);
    float value = fabs(sin(theta))+1e-6;
    for (int x=0; x < width; x++) {
      float v1 = fabs(sin(20*float(x)/width));
      float v2 = fabs(sin(21*float(x)/width));
      float v3 = fabs(sin(theta+x/width*14.f)+0.1f);
      int ww = (width * y + x)*3;
      // set it to the output format as required by RGBE_WRITE
      cols[ww+0] = r*value*v1;
      cols[ww+1] = g*value*v2;
      cols[ww+2] = b*value*v3;
    } // for jcol
  } // for irow
}

float
CEnvironmentMap::computeProb(float x, float y, float theta, float RGB[3]) const
{
#if 1
  float prob = luminance(RGB) * totalPowerInv;
#else  
  // According to Pharr/Humphreys2004 the absolute value of the Jacobian
  // 2*PI*PI from rectangle to rectangle and further Jacobian sin(theta)*R*R
  // applies for conversion between theta-phi and sphere.
#ifdef FIRST_ROWS
  float sinTheta = sin(theta)+1e-31;
  assert(sinTheta > 0.f);
  int iy = (int)y;
  float betaY = y-(float)iy;
  int ix = (int)x;
  int ix2 = (ix+1)%width;
  float alphaX = x-(float)ix;
  float fprob[2];
  int iy2 = iy+1;
  if (iy2 >= height) iy2 = height-1;

  // in X - interpolation between two rows
  fprob[0] = float(width)*
    ((lums[iy*width+ix]*(1.f-alphaX) + lums[iy*width+ix2]*alphaX)*(1.f-betaY) +
     (lums[iy2*width+ix]*(1.f-alphaX) + lums[iy2*width+ix2]*alphaX)*betaY)
    / (cumlums[addrcum(width-1,iy)]*(1.f-betaY) + cumlums[addrcum(width-1,iy2)]*betaY);

  // in Y - interpolation between two columns
  fprob[1] = float(height)*(cumlums[addrcum(width-1, iy)]*(1.f-betaY) + cumlums[addrcum(width-1, iy2)]*betaY)/
    cumrows[height-1];
  //cout << fprob[0] << " " << fprob[1] << endl;

  float prob = fprob[0]*fprob[1]/(2.0*sin(theta));    
#else  // FIRST_ROWS
  // as in PBRT - the columns are integrated first after multiplication by sin(theta)
  // the integrated columns are added to a single row
  float sinTheta = sin(theta)+1e-31;
  assert(sinTheta > 0.f);
  int iy = (int)y;
  int iy2 = iy+1;
  if (iy2 >= height-1) iy2=height-1;
  float betaY = y-(float)iy;
  int ix = (int)x;
  int ix2 = (ix+1)%width;
  float alphaX = x-(float)ix;
  float fprob[2];
  // in X - interpolation between two columns from cumcols - this is easy
  fprob[0] = float(width)*(cumlums[addrcum(ix, height-1)]*(1.f-alphaX) + cumlums[addrcum(ix2, height-1)]*alphaX)/
    cumcols[width-1];

  // in Y - interpolation between two columns - marginal probability interpolated
  fprob[1] = float(height) *
    ((lums[iy*width+ix]*(1.f-alphaX) + lums[iy*width+ix2]*alphaX)*(1.f-betaY) +
     (lums[iy2*width+ix]*(1.f-alphaX) + lums[iy2*width+ix2]*alphaX)*betaY)
    / (cumlums[addrcum(ix,height-1)]*(1.f-alphaX) + cumlums[addrcum(ix2,height-1)]*alphaX);
  //cout << fprob[0] << " " << fprob[1] << endl;
  float prob = fprob[0]*fprob[1]/(2.f*sinTheta);
  //cout << "prob = " << prob << " theta = " << theta << endl;
#endif // FIRST_ROWS
#endif
  return prob;
}

// get value RGB for a particular theta and phi, when we assume
// that the HDR image is the environment map
bool 
CEnvironmentMap::getRGB_ThetaPhi(float theta, float phi, float RGB[3]) const
{
  assert(0.f <= phi);
  assert(phi <= 2.f*M_PI);
  assert(0.f <= theta);
  assert(theta <= M_PI);
  float X, Y;
  X = phiToX(phi);
  Y = thetaToY(theta);
  int xx = X;
  int xx2 = (xx+1)%width;
  float alpha = X - (float)xx;
  int yy = Y;
  assert(yy < height);
  int yy2 = yy+1;
  if (yy2 >= height) yy2 = yy;
  float beta = Y - (float)yy;
  // interpolated value for all 3 channels
  for (int isp = 0; isp < 3; isp++) {
    RGB[isp] = getValueChannel(xx, yy, isp) * (1.f-alpha) * (1.f-beta) +
      getValueChannel(xx2, yy, isp) * alpha * (1.f-beta) +
      getValueChannel(xx, yy2, isp) * (1.f-alpha) * beta +
      getValueChannel(xx2, yy2, isp) * alpha * beta;
  } // for isp
  //it is not checked anywhere I assume false = SUCCESS ?
  return false;
}

// get value RGB for a particular theta and phi, when we assume
// that the HDR image is the environment map
float
CEnvironmentMap::getRGBandProb_ThetaPhi(float theta, float phi, float RGB[3]) const
{
  float X, Y;
  X = phiToX(phi);
  assert(0.f <= phi);
  assert(phi <= 2.f*M_PI);
  Y = thetaToY(theta);
  assert(0.f <= theta);
  assert(theta <= M_PI);
  int xx = X;
  int xx2 = (xx+1)%width;
  float alpha = X - (float)xx;
  int yy = Y;
  assert(yy < height);
  int yy2 = yy+1;
  if (yy2 >= height) yy2 = yy;
  float beta = Y - (float)yy;
  // interpolated value for all 3 channels
  for (int isp = 0; isp < 3; isp++) {
    RGB[isp] = getValueChannel(xx, yy, isp) * (1.f-alpha) * (1.f-beta) +
      getValueChannel(xx2, yy, isp) * alpha * (1.f-beta) +
      getValueChannel(xx, yy2, isp) * (1.f-alpha) * beta +
      getValueChannel(xx2, yy2, isp) * alpha * beta;
  } // for isp

  float prob = computeProb(X, Y, theta, RGB);  
  return prob;
}

// Computes the statistics over north hemisphere
void
CEnvironmentMap::computeMeanAndVarianceHemisphere(float normalXYZ[3], float &meanHS, float &varianceHS, int cntSamples)
{
  // hemisphere has defined center by normal 'normalXYZ[3]
  assert(vec_norm(normalXYZ));
  // we use cosine sampling over hemisphere

  // and incremental algorithm to compute a variance
  // create coordinate system given by normal, T, B
  float Tint[3], Bint[3];
  ArbitraryNormal(normalXYZ, Tint); // generate arbitrary vector T perpendicular to normal
  vec_crossprod(normalXYZ, Tint, Bint); // and then B to create the coordinate system
  //  SVAR sX, sY, sZ, varmean2;
  //long double sum = 0., sum2 = 0.;
  SVAR varmean;
  
  // We compute not so many samples, it is only initial approximation
  for (int i = 0; i < cntSamples; i++) {
    // we compute 1/PI sampling over hemisphere here, see global illumination compendium
    float rndxx = RandomValue();
    float rndyy = RandomValue();
    float sin_theta = sqrt(1.0000 - rndxx * rndxx);
    if (sin_theta >= 1.0f) sin_theta=0.99999;
    float phi      = 2.0 * M_PI * rndyy;
    float XYZL[3];
#ifdef XYZCOORSYSTEM
    // Z-axis is in the direction of north-pole
    // plane XY corresponds to theta=PI/2 and phi<0,2PI)
    const int Xi=0, Yi = 1, Zi = 2;
#else
    // Y-axis is in the direction of north-pole
    // plane XZ corresponds to theta=PI/2 and phi<0,2PI)
    const int Xi=0, Yi = 2, Zi = 1;
#endif
    XYZL[Xi] = cos(phi) * sin_theta;
    XYZL[Yi] = sin(phi) * sin_theta;
    XYZL[Zi] = rndxx; // cos(theta)
    
    float XYZ[3]; // direction in world-coordinate system
    // This is transpose matrix from local coordinate system to world-space system
    // it is possibly only since vectors T,B,N create orthogonal matrix, then inverse
    // matrix is equal to the transposed matrix
    XYZ[0] = Tint[0]*XYZL[0] + Bint[0]*XYZL[1] + normalXYZ[0]*XYZL[2];
    XYZ[1] = Tint[1]*XYZL[0] + Bint[1]*XYZL[1] + normalXYZ[1]*XYZL[2];
    XYZ[2] = Tint[2]*XYZL[0] + Bint[2]*XYZL[1] + normalXYZ[2]*XYZL[2];
    // For sure compute dotprod ant check it
    vec_norm(XYZ);
    // cosine(theta) in local coordinate system given by NormalXYZ, Tint, Bint
    float cosTheta = vec_dotprod(XYZ,normalXYZ);
    if ((cosTheta < -1e-3f)||(fabsf(XYZL[2] - cosTheta) > 1e-3f)) {
      cout << "ERROR - generated directions is wrong" << endl;
      cout << endl;
      abort();
    }
    cosTheta = XYZL[2];
    
    // now let us access the value
    float theta2, phi2;
    ConvertXYZtoThetaPhi(XYZ, theta2, phi2);
#if 1
    float RGB[3];
    this->getRGB_ThetaPhi(theta2, phi2, RGB);
    float lum = luminance(RGB);
#else
    float lum = this->getLUM_DIRapprox(theta2, phi2);
    //float lum = this->getLUM_DIRexact(theta2, phi2);
#endif
    //    cout << "lum = " << lum << " ";
    if ((lum < 0.f)||(isnan_local(lum))) {
      cout << "Problem = " << lum << endl;
      abort();
    }
    varmean.update(lum);
#if 0
    float lum2 = lum / cosTheta;
    sum += lum;
    sum2 += lum*lum;
    sX.update(XYZ[0]);
    sY.update(XYZ[0]);
    sZ.update(XYZ[0]);
    varmean2.update(lum2);
#endif    
  } // for all generated directions

  // we have estimated both variance and mean
  double varianceHemisphere =  varmean.evaluate();
  //double variance2H = varmean.evaluate();
#if 0
  cout << "DIR = " << normalXYZ[0] << " " << normalXYZ[1] << " " << normalXYZ[2] << endl;
  cout << "sum = " << sum << " sum2 = " << sum2 << " VARfromSUMS = " << sum2 - sum*sum
       << " sxV = " << sX.evaluate() << " syV = " << sY.evaluate() << " szV = " << sZ.evaluate()
       << " varmean.mean = " << varmean.mean << " varmean2.mean = " << varmean2.mean
       << " varmean.var = " << varmean.evaluate() << " varmean2.var = " << varmean2.evaluate() << endl;
#endif  
  //  cout << "varianceHemisphere = " << varianceHemisphere; // << " mean = " << mean;
  //  cout << endl;

  // copy the data to the output
  meanHS = varmean.mean * 2.0f * M_PI; // should be normalized by sphere surface to get 1.0 for cosine function and 2*PI for constatnt function
  varianceHS = varmean.evaluate() * 4.0 * M_PI * M_PI; // also variance is increased by the same factor squared
  return;
}

// coordinates of one of the icosahedron vertex needs the two constants
#define XF 0.525731112119133696
#define ZF 0.850650808352039932

// icosahedron  vertices
static float icosahedronvertex[12][3] = {
  {-XF, 0.0, ZF}, {XF, 0.0, ZF}, {-XF, 0.0, -ZF}, {XF, 0.0, -ZF},
  {0.0, ZF, XF}, {0.0, ZF, -XF}, {0.0, -ZF, XF}, {0.0, -ZF, -XF},
  {ZF, XF, 0.0}, {-ZF, XF, 0.0}, {ZF, -XF, 0.0}, {-ZF, -XF, 0.0}
};

// icosehedron faces
static int icosahedrontriangle[20][3] = {
  {1,4,0}, {4,9,0}, {4,5,9}, {8,5,4}, {1,8,4},
  {1,10,8}, {10,3,8}, {8,3,5}, {3,2,5}, {3,7,2},
  {3,10,7}, {10,6,7}, {6,11,7}, {6,0,11}, {6,1,0},
  {10,1,6}, {11,0,9}, {2,11,9}, {5,2,9}, {11,2,7}
};

// Here we insert basic directions based on the icoseheadron subdivision of sphere
void
CEnvironmentMap::insertInitialSphereDirSamples()
{
  // for all faces
  for(int i = 0; i < 20; i++) {
    // compute the center of a triangle
    float dXYZ[3];
    // for all xyz of coordinate system for vectors
    for (int j=0; j<3;j++) {
      dXYZ[j] = 0; // sum
      // for all 3 vertices of a triangle
      for (int k=0; k<3;k++) { // three different vertices
	int vertexIndex = icosahedrontriangle[i][k];
	dXYZ[j] += icosahedronvertex[vertexIndex][j];
      } // for 3 vertices of a triangle
    } // for j
    // now we have a center of a face 'i'
    vec_norm(dXYZ);
    float mean, var;
    getMeanVarianceHemisphere(dXYZ, mean, var, true); // enforce to compute variance and mean value
  } // for i
  for (int i = 0; i < 12; i++) {
    //take individual 12 vertices as well
    float mean, var;
    getMeanVarianceHemisphere(icosahedronvertex[i], mean, var, true); // enforce to compute variance
  }
  return;
}

void
CEnvironmentMap::getMeanVarianceHemisphere(float dirXYZ[3], float &meanHemisphere,
					   float &varianceHemisphere, bool computeNew)
{
  // normalize the direction
  vec_norm(dirXYZ);
  const float diffCosineMax = cos(15.f*M_PI/180.f); //max difference in degrees converted to radiance
  if (cntDirsVDmax == 0) {
    // This only generates direction
    cntDirsVDcur = 0;
    // estimate number of directions for whole sphere, this is only approximation
    // with some safety factor SF.
    float diffSin = sqrt(1.0f - sqr(diffCosineMax));
    //PI*r*r is the surface area of one sphere on the surface of sphere
    float SAoneCircle = M_PI*sqr(diffSin); 
    // surface area of sphere, divided by sphere of one circle, multiplied by safety factor for sure
    const float SF = 3.1f; // SF factor 2.0f is good for 16 degrees above
    cntDirsVDmax = (int)(4.0f*M_PI/diffSin * SF);
    // cout << " cntDirsVDmax = " << cntDirsVDmax << endl;
    meanAndVarPartialDirVal = new float[cntDirsVDmax*6]; // one value reserved
    assert(meanAndVarPartialDirVal);
    cntDirsVDcur = 0;
    // insert initial samples given by subdivision of sphere with icosahedron
    insertInitialSphereDirSamples();
  }
  if (!computeNew) {
    // we use a simple linear search, which is fast enough for our application
    // if not, it should be sped up somehow.
    // the data is allocated, find out the closest and compute angle
    // between already computed direction and query vector
    for (int i = 0; i < cntDirsVDcur; i++) {
      float cosAngle = vec_dotprod(&(meanAndVarPartialDirVal[6*i]), dirXYZ);
      if (cosAngle >= diffCosineMax) {
	// return precomputed value, first one found, we are finished
	//  cout << "i=" << i << endl;
	meanHemisphere = meanAndVarPartialDirVal[6*i+3];
	varianceHemisphere = meanAndVarPartialDirVal[6*i+4];
	return; // we are finished
      }
    } // for
  }
  // value requested to be computed or 
  // value not found, precompute it for a given direction dirXYZ
  for (int isp = 0; isp < 3; isp++) // insert the direction to the array
    meanAndVarPartialDirVal[6*cntDirsVDcur+isp] = dirXYZ[isp];
  // compute the result to the output variables
  computeMeanAndVarianceHemisphere(dirXYZ, meanHemisphere, varianceHemisphere);
  // and store it as well
  meanAndVarPartialDirVal[6*cntDirsVDcur+3] = meanHemisphere;
  meanAndVarPartialDirVal[6*cntDirsVDcur+4] = varianceHemisphere;
  //cout << "M = " << meanHemisphere << " V = " << varianceHemisphere << endl << flush;  
  cntDirsVDcur++; // the next storage
  //cout << "cntDirsVDcur = " << cntDirsVDcur << endl;

  if (cntDirsVDcur == cntDirsVDmax) {
    cout << "Increase the constant SF in the code hdrloader.cpp in function getVarianceHemisphere(float dirXYZ[3])"
	 << " line = " << __LINE__ << " file = " << __FILE__ << endl;
    abort();
  }

  return;
}



