// ===================================================================
// $Id: $
//
// hdrloader.h
//
// Class: CHDRImage, CEnvironmentMap
//
//  The declaration of the class for reading HDR image and environment maps
//
// Initial coding by Vlastimil Havran, 2013.

#ifndef __HDRLOADER_H__
#define __HDRLOADER_H__

#include <cassert>
#include "AuxRND.h"
#include "tgaimg.h"

// This is the class for environment map
class CHDRImage
{
protected:
  int width, height;
  // each pixel takes 3 float32, each component can be of any value...
  float *cols;
  bool typeHDR;
  string filenameRead; // if read, then filename
  string gname; // the name without extension and path
  void extractName(const char *fileName);
public:
  CHDRImage():width(0), height(0), cols(0), typeHDR(false) { }
  CHDRImage(const CHDRImage &src); // copy constructor
  // copy constructor with possibly making bargraph on the right
  CHDRImage(const CHDRImage &src, int pseudoColorTechnique,
	    bool copyImage = true, int widthBargraph = 30, int borderBargraph = 5);

  ~CHDRImage() { delete []cols;}

  // Read the data from HDR file

  // returns false on error
  bool loadHDR(const char *fileName);
  bool loadPNG(const char *fileName);
  bool saveHDR(char *fileName, bool verbose = false);
  bool savePNG(const char *fileName);
  bool getTypeHDR() const { return typeHDR;}
  void copyDataFromArray(int nwidth, int nheight, int channels,
			 float *buffer, int bufferWidth,
			 int xoffset, int yoffset,
			 bool flipVertically=false);
  virtual void allocate(int nwidth, int nheight, bool ntypeHDR=false) {
    delete []cols;
    width = nwidth;
    height = nheight;
    typeHDR = ntypeHDR;
    int nmax = 3 * height * width;
    cols = new float[nmax];
    assert(cols);
    for (int i = 0; i < nmax; i++) cols[i] = 0;
  }
  int getWidth() const { return width;}
  int getHeight() const { return height;}
  
  const string getName() const {return gname;}
  const string getFileName() const {return filenameRead;}
  
  // get value RGB for a particular theta and phi, when we assume
  // that the HDR image is the environment map
  void setValueChannel(int x, int y, int isp, float value) {
    assert((x >= 0)&&(x < width));
    assert((y >= 0)&&(y < height));
    cols[(width * y + x)*3 + isp] = value;
  }
  // get value RGB for a particular theta and phi, when we assume
  // that the HDR image is the environment map
  float getValueChannel(int x, int y, int isp) const {
    assert((x >= 0)&&(x < width));
    assert((y >= 0)&&(y < height));
    return cols[(width * y + x)*3 + isp];
  }
  
  // get the color when indexed as image
  bool getRGB_XY(int x, int y, float RGB[3]) const;
  // access directly to the triple of floats
  float* getRGB_XY(int x, int y) {return cols+3*(width*y+x);}
  // raw access to the data from beginning
  float *getBuf() {return cols;}

  // get the color when indexed as image
  bool setRGB_XY(int x, int y, float r, float g, float b);

  // make a mark at specific point in the image
  void makeMarkRGB(int mx, int my, EMarkType typeMark,
		   int marksize = 2, float mult = 1.0f);
  
  // draw a line between two points, no clipping possible
  void drawLine(int x1, int y1, int x2, int y2,
		float red, float green, float blue);
  // the whole image is filled in by the same color
  void setConstant(float r, float g, float b);

  // shifts circularly the image to left or right
  void horizontallyShift(int pixelOffset);

  void Multiply(float coeff, int x1=0, int y1=0, int x2=-1, int y2=-1);
  void AddOffset(float coeff, int x1=0, int y1=0, int x2=-1, int y2=-1);
  void Minimize(float coeff, int x1=0, int y1=0, int x2=-1, int y2=-1);
  void Maximize(float coeff, int x1=0, int y1=0, int x2=-1, int y2=-1);
  void Fabs(int x1=0, int y1=0, int x2=-1, int y2=-1);
  void MakeGray(int x1, int y1, int x2, int y2);
  void MakeWholeGray();

  // Compare two images and prints possibly the output
  bool CompareTo(CHDRImage &i2, int verbose) const;
  // Substruct two images to the output file, possibly given by the user
  CHDRImage* Substruct(CHDRImage &i2, CHDRImage *i3 = 0,
		       bool computeRMSE = true, int verbose = true) const;
  // Set pseudocolor from image to the output i2 image
  int setsPseudocolor(CHDRImage &i2, int pseudoColorTechnique,
		      float lowerBound, float upperBound) const;  
};


// get the color when indexed as image
inline bool
CHDRImage::getRGB_XY(int x, int y, float RGB[3]) const
{
  if ((x < 0)||(x >= width)) return true; // error
  if ((y < 0)||(y >= height)) return true; // error
  int itt = (width*y+x)*3;
  RGB[0] = cols[itt];
  RGB[1] = cols[itt+1];
  RGB[2] = cols[itt+2];
  return false;
}

// get the color when indexed as image
inline bool
CHDRImage::setRGB_XY(int x, int y, float r, float g, float b)
{
  if ((x < 0)||(x >= width)) return true; // error
  if ((y < 0)||(y >= height)) return true; // error
  int itt = (width*y+x)*3;
  cols[itt] = r;
  cols[itt+1] = g;
  cols[itt+2] = b;
  return false;
}

// -------------------------------------------------------------------
class CEnvironmentMap:
  public CHDRImage
{
  // only luminance when computed float32
  float *lums;
  // luminance in cummulative sum for inverse transform method
  double *cumlums;
  // luminance in cummulative sum for inverse transform method
//#define FIRST_ROWS
#ifdef FIRST_ROWS
  // the first implementation, first summing along rows, then along columns
  double *cumrows;
  // addressing CDF
  int addrcum(int x, int y) const {return y*width+x;}
  // addressing PDF
  int addrlum(int x, int y) const {return y*width+x;}
#else // FIRST_ROWS
  // second impl, as in PBRT book, first summing in columns, then in rows
  double *cumcols;
  // addressing CDF
  int addrcum(int x, int y) const {return x*height+y;}
  // addressing PDF
  int addrlum(int x, int y) const {return y*width+x;}
#endif // FIRST_ROWS
  // the total power integrated over the sphere in the image
  double totalPower, totalPowerOrig;
  float totalPowerInv;
  // positions and colours of precomputed lights
  float **envLightPositions;
  float **envLightColours;
  float **envLightPositionsBackup;
  float **envLightColoursBackup;
  float *envLightProbs;
  int cntAllocatedLights; // the number of allocated/precomputed lights
  int cntLights; // the number of lights
  // computation of variance
  float varianceWholeSphere;
  int   cntDirsVDmax; // maximum number of directions
  int   cntDirsVDcur; // current number of directions
  float *meanAndVarPartialDirVal;
  // if to randomize selection from precomputed lights
  bool randomizePrecomputedLights;
  int randomizedIndexLight;
  // local functions
  float computeProb(float x, float y, float theta, float RGB[3]) const;  
  void deallocatePrecomputedLights();
  void allocatePrecomputedLights(int cnt);
  inline bool setLUM_XY(int x, int y, float val);
  void insertInitialSphereDirSamples(); // initial directions on the sphere
public:
  CEnvironmentMap();
  CEnvironmentMap(const CEnvironmentMap &src); // copy constructor
  ~CEnvironmentMap();

// If this macro is defined, then use classic coordinate system XYZ, where
// Z-axis is in the direction of north pole
// plane XY corresponds to theta=PI/2 and phi <0, 2PI)
// If macro is undefined, use classic coordinate system XYZ, but the north pole
// is in the direction of Y-axis 
#define XYZCOORSYSTEM
  
  // This is for vizualization
  //  bool saveDebugTGA(const char *fileName, int mode, float mult = 1.0);
  bool saveDebugHDR(char *fileName, int mode, float mult = 1.0);

  // Allocate the image for manual filling in
  void allocate(int nwidth, int nheight, bool ntypeHDR=false) {
    delete []lums; lums = 0;
    delete []cumlums; cumlums = 0;
#ifdef FIRST_ROWS
    delete []cumrows; cumrows = 0;
#else
    delete []cumcols; cumcols = 0;
#endif
    // Allocate only color values
    CHDRImage::allocate(nwidth, nheight, ntypeHDR);
  }

  // Compute luminance from color values
  void computeLuminance();
  // Normalize the power of EM map so the integral is 1 Watt.
  void normalizePower();
  // debug function
  bool computeCDF();
  // computes the total power
  void computeTotalPower();
  // returns total power
  double getTotalPower() const { return totalPower;} 
  // returns total power original without possible normalization
  double getTotalPowerOrig() const { return totalPowerOrig;}

  // get value RGB for a particular theta and phi, when we assume
  // that the HDR image is the environment map
  bool getRGB_ThetaPhi(float theta, float phi, float RGB[3]) const;
  // get value probability for a particular theta and phi
  float getProb_ThetaPhi(float theta, float phi) const;
  // This returns probability and sets RGB together
  float getRGBandProb_ThetaPhi(float theta, float phi, float RGB[3]) const;
  
  // get luminance when indexed as image
  bool getLUM_XY(int x, int y, float &lum) const;
  float getLUM_DIRexact(float theta, float phi) const;
  float getLUM_DIRapprox(float theta, float phi) const;
  // get cummulative sum of luminance when indexed as image by X and Y coordinates
  bool getCUMLUM_XY(int x, int y, float &clum) const;  

  // For two random values rnd1 and rnd2 compute X-Y position on the
  // image and set the output RGB value
  bool importanceSamplingXY(float rnd1, float rnd2, 
			    float &X, float &Y, float RGB[3]) const;

  // For two random values rnd1 and rnd2 compute theta-phi position on the
  // full sphere and set the output RGB value
  bool importanceSamplingThetaPhi(float rnd1, float rnd2,  // input
				  float &theta, float &phi, // output
				  float RGB[3], float &prob) const; // output

  float thetaToY(float theta) const {
    assert(0 <= theta); assert(theta <= M_PI);
    // bottom line (0) of the image is PI, (height-1)-line is then 0
    return (height-0.9999f)*theta/M_PI;
  }
  float yToTheta(float y) const {
    assert(0 <= y); assert(y < height);
    return y/(height-0.9999f) * M_PI; // bottom line (0) of the image is PI
  }
  float phiToX(float phi) const {
    assert(phi >= 0); assert(phi<=2.f*M_PI);
    return phi/(2.f*M_PI) * (width-0.9999f);
  }
  float xToPhi(float x) const {
    assert(x >= 0); assert(x < width);
    return (x*2.f*M_PI)/(width-0.999f);
  }
  // precompute cnt directional lights to approximate environment map
  void precomputeLights(int cnt);
  // how many lights are currently precomputed
  int  countPrecomputedLights() const { return cntLights;}

  // get current mode if randomization applies
  bool getModeRandomLights() const { return randomizePrecomputedLights;}
  // set random access to precomputed lights
  void setModeRandomLights(bool randomFlag) {
    randomizePrecomputedLights = randomFlag;
    if (randomFlag == true) {
      const int thresholdLights = 10000;
      if (cntLights < thresholdLights) {
	cout << "ERROR: For randomization of deterministic lights you have"
	     <<" to increase the number of lights to at least "
	     << thresholdLights << endl;
	abort();
      }
    }
  }
  void generateNewRandomLightPrecomputed() {
    if (randomizePrecomputedLights) {
      randomizedIndexLight = RandomValue() * cntLights;
      assert(randomizedIndexLight>=0); assert(randomizedIndexLight<cntLights);
    }
  }
  // Accessing functions for direction for precomputed lights
  float getDirectionPrecomputedLight(int i, int ixyz) const {
    if (randomizePrecomputedLights) { i = randomizedIndexLight;}
    assert(i>=0); assert(i<cntLights);
    assert(ixyz>=0); assert(ixyz<3);
    return envLightPositions[i][ixyz];
  }
  float* getDirectionPrecomputedLight(int i) {
    if (randomizePrecomputedLights) { i = randomizedIndexLight;}
    assert(i>=0); assert(i<cntLights);
    return envLightPositions[i];
  }
  void setDirectionPrecomputedLight(int i, float XYZ[3]) {
    if (randomizePrecomputedLights) { i = randomizedIndexLight;}
    assert(i>=0); assert(i<cntLights);
    for (int isp=0; isp<3;isp++) envLightPositions[i][isp]=XYZ[isp];
  }

  // Accessing functions for colors for precomputed lights
  float getColorPrecomputedLight(int i, int irgb) const {
    if (randomizePrecomputedLights) { i = randomizedIndexLight;}
    assert(i>=0); assert(i<cntLights);
    assert(irgb>=0); assert(irgb<3);
    return envLightColours[i][irgb];
  }
  float* getColorPrecomputedLight(int i) {
    if (randomizePrecomputedLights) { i = randomizedIndexLight;}
    assert(i>=0); assert(i<cntLights);
    return envLightColours[i];
  }
  void setColorPrecomputedLight(int i, float RGB[3]) {
    if (randomizePrecomputedLights) { i = randomizedIndexLight;}
    assert(i>=0); assert(i<cntLights);
    for (int isp=0; isp<3;isp++) envLightColours[i][isp]=RGB[3];
  }

  // returns value of probability density function for a precompute light 'i'
  float getProbPrecomputedLight(int i) const {
    if (randomizePrecomputedLights) { i = randomizedIndexLight;}
    assert(i>=0); assert(i<cntAllocatedLights);
    return envLightProbs[i];
  }
  void makeLighter(int i, float c1, float c2);
  void restoreColorFromBackup(int i) { assert(i >= 0); assert(i<cntAllocatedLights);
    for(int j=0; j < 3; j++) envLightColours[i][j] = envLightColoursBackup[i][j];
  }
  // rotates only precomputed lights around y-axis
  void rotateEnvironmentLights(float angle);
  // make mark in luminance channel in the image
  void makeMarkLUM(int mx, int my, EMarkType typeMark, int marksize = 2,
		   float mult = 1.0f);

  // for debugging only
  // North Hemisphere
  void setFunctionConstNorthHemisphere(float r, float g, float b);
  void setFunctionCosineTheta(float r, float g, float b);
  void setFunctionCosineThetaNorthHemisphere(float r, float g, float b);
  // South hemisphere
  void setFunctionConstSouthHemisphere(float r, float g, float b);

  // Some other testing functions setting the data
  void setFunction2(float r, float g, float b);
  void setFunction3(float r, float g, float b);
  
  // --------------------------------------------------------
  // variance support
  // return mean and variance for some direction - this is only an approximation
  // (northpole has [0,0,1] or [0, 1, 0] coordinates)
  void getMeanVarianceHemisphere(float dirXYZ[3], float &meanHemisphere,
				 float &varianceHemisphere, bool computeNew=false);
  // computing the statistical values for some given vector
  // (northpole has [0,0,1] or [0, 1, 0] coordinates)
  void computeMeanAndVarianceHemisphere(float XYZ[3], float &meanHS, float &varHS, int cntSamples = 1000);
};

// ---------------------------------------------------------------------------
// Inline methods


// get luminance when indexed as image
inline bool
CEnvironmentMap::getLUM_XY(int x, int y, float &lum) const
{
  if ((x < 0)||(x >= width)) return true; // error
  if ((y < 0)||(y >= height)) return true; // error
  lum = lums[width*y+x];
  return false;
}

inline float
CEnvironmentMap::getLUM_DIRexact(float theta, float phi) const
{
  float yf = thetaToY(theta);
  int iy = (int)yf;
  assert(yf>=0); assert(yf<height);
  int iy2 = iy+1;
  if (iy2 >=height) iy2 = height-1;
  float beta = yf - iy;

  float xf = phiToX(phi);
  assert(xf>=0); assert(xf<width);
  int ix = (int)ix;
  int ix2 = (ix+1)%width;
  float alpha = xf - ix;
  float lumsExact = (lums[width*iy+ix]*(1.-alpha)+lums[width*iy+ix2]*alpha)*(1.-beta)+
    (lums[width*iy2+ix]*(1.-alpha)+lums[width*iy2+ix2]*alpha)*(1.-beta);
  return lumsExact;
}

inline float
CEnvironmentMap::getLUM_DIRapprox(float theta, float phi) const
{
  int iy = (thetaToY(theta)+0.5f); // rounded to the first coordinate
  if (iy >= height) iy=height-1;
  int ix = ((int)(phiToX(phi)+0.5))%width;
  assert(iy>=0); assert(iy<height);
  assert(ix>=0); assert(ix<width);
  return lums[width*iy+ix]; // nearest neighbor value
}

// get cummulative sum of luminance when indexed as image
inline bool
CEnvironmentMap::getCUMLUM_XY(int x, int y, float &clum) const
{
  if ((x < 0)||(x >= width)) return true; // error
  if ((y < 0)||(y >= height)) return true; // error
  clum = cumlums[width*y+x];
  return false;
}

#endif // __HDRLOADER_H__
