// ================================================================
// $Id: $
//
// render.h
//
// Header file to declare basic rendering functions given a point in 3D space
//
// Initial coding by Vlastimil Havran, 2013.

#ifndef __RENDER_H__
#define __RENDER_H__

#include "configh.h"
#include "brdfmodel.h"
#include "tangles.h"
#include "hdrloader.h"
#include "lightenv.h"

// Counters and information for debugging purposes
extern int debugPosX;
extern int debugPosY;
// global counters for whole image
extern long long int failBRDFcnt;
extern long long int evalBRDFcnt;
extern long long int failEMcnt;
extern long long int evalEMcnt;
// the counters of winning for whole image
extern long int winCntVX; // = BRDF
extern long int winCntVY; // = EM
extern long int evalCntVXY;
// the counters of usage EM or BRDF sampling in a single text
extern int evalEMcntST;
extern int evalBRDFcntST;
const float  minValueVIS = 0;
const float  maxValueVIS = 1;
extern float evalValueVIS;
inline void resetCountersST() { evalEMcntST = evalBRDFcntST = 0;}
inline void setCountersST(int BRDF, int EM) { evalBRDFcntST=BRDF; evalEMcntST=EM; }
inline void setValueST(float vv) { evalValueVIS=vv;}

// This implements one-point-light illumination
void         // interpolated vertices, tangents, normals and binormals
computeMode0(float Vint[3], float Tint[3], float Nint[3], float Bint[3], 
	     // theta of viewer, phi of viewer, texture coords, index to data, and the intensity multiplier
	     float t_v, float p_v, int irow, int jcol, int BTFindex, float normF); 

// This implements deterministic sampling from environment map, all pixels the same set of lights
void        // interpolated vertices, tangents, normals and binormals
computeMode1(float Vint[3], float Tint[3], float Nint[3], float Bint[3], 
	     // theta of viewer, phi of viewer, texture coords, index to data, and the intensity multiplier
	     float t_v, float p_v, int irow, int jcol, int BTFindex, float normF); 

// This implements random sampling from environment map, each pixel different set of light sources (random)
void        // interpolated vertices, tangents, normals and binormals
computeMode2(float Vint[3], float Tint[3], float Nint[3], float Bint[3], 
	     // theta of viewer, phi of viewer, texture coords, index to data, and the intensity multiplier
	     float t_v, float p_v, int irow, int jcol, int BTFindex, float normF);

// This implements random sampling from BRDF*cos(theta)
void        // interpolated vertices, tangents, normals and binormals
computeMode3(float Vint[3], float Tint[3], float Nint[3], float Bint[3], 
	     // theta of viewer, phi of viewer, texture coords, index to data, and the intensity multiplier
	     float t_v, float p_v, int irow, int jcol, int BTFindex, float normF); 

// This implements MIS, combining EM and BRDF*cos(theta) sampling, without storing the samples
void         // interpolated vertices, tangents, normals and binormals
computeMode4(float Vint[3], float Tint[3], float Nint[3], float Bint[3],
	     // theta of viewer, phi of viewer, texture coords, index to data, and the intensity multiplier
	     float t_v, float p_v, int irow, int jcol, int BTFindex, float normF);

// This is different implementation of MIS, combining EM and BRDF*cos(theta) sampling, with storing the samples
void         // interpolated vertices, tangents, normals and binormals
computeMode5(float Vint[3], float Tint[3], float Nint[3], float Bint[3],
	     // theta of viewer, phi of viewer, texture coords, index to data, and the intensity multiplier
	     float t_v, float p_v, int irow, int jcol, int BTFindex, float normF);

inline float
evaluateMISweight(float prob1, int cntSamples1, float prob2, int cntSamples2)
{
  const int modeHeuristics = 0;
  if (modeHeuristics == 0) {
    // balanced heuristic is here
    float contrib1 = (float)cntSamples1*prob1;
    return contrib1/(contrib1 + (float)cntSamples2*prob2);    
  }
  if (modeHeuristics == 1) {
    // power heuristics
    const float alphaCoeff = 2.0f;
    float contrib1 = powf((float)cntSamples1*prob1, alphaCoeff);
    return contrib1/(contrib1 + powf((float)cntSamples2*prob2, alphaCoeff));
  }
  if (modeHeuristics == 2) {
    // cut-off heuristic with some threshold
    const float alphaCoeff = 0.2f; // minimum threshold
    float contrib1 = cntSamples1*prob1;
    float contrib2 = cntSamples2*prob2;
    if (contrib1/(contrib1+contrib2) < alphaCoeff)
      return 0; // the first sampling technique does not have enough weight
    if (contrib2/(contrib1+contrib2) < alphaCoeff)
      return 1; // the second sampling technique does not have enough weight
    return contrib1/(contrib1+contrib2);
  }
  //  cout << "Error - wrong MIS model used << endl;
}

// Entry to store result for MIS evaluation later on
struct SEntry {
  float RGB[3];
  float cosTheta; // cosine(theta) for this sample with respect to normal
  float probThis; // probabilty of sampling the RGB value above from first technique
  float probOther;// probabilty of sampling the RGB value above from second technique
  int technique; // the technique indicator .. = 0 - first technique .. = 1 - second technique
  int padding;
  void setZero() { RGB[0] = 0; RGB[1] = 0; RGB[2] = 0; cosTheta=1.f;probThis=probOther=0; technique=-1;}
};

// For BTF data max size of texture
#define REPEATX 512
#define REPEATY 512

// the pixel to debug 68/245
const int irowCheck = -124;
const int jcolCheck = 278;

#define MARK_PROBLEM 1e6

#endif // __RENDER_H__
