// ===================================================================
// $Id: $
//
// tangles.cpp
//
//  The class to implement rendering via OpenGL with shaders and HDR texture
//
// Initial coding by Jiri Filip, 2006.

#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <cassert>
#include <string>

#include "bconfig.h"
#include "AuxFuncs.h"
#include "TAlloc.h"
#include "tangles.h"

//#define FLOATMAX ((float)(1e30))
#define EPS 3.0e-7
#define INTMAX 99999999
#define PI 3.14159265358979323846


//#########################################################################################
//# Class TAngles                                                                         #
//#########################################################################################

TAngles::TAngles()
{
  NoOfDirections = 81;
  angles = allocation2(0,NoOfDirections,1,7);

  c_pos = allocation1(0,2);
  l_pos = allocation1(0,2);

  i_weight = allocation1(0,2);
  v_weight = allocation1(0,2);
  i_index = iallocation1(0,2);
  v_index = iallocation1(0,2);

  texWeights = allocation1(0,8);
  illuPos = iallocation1(0,8);
  viewPos = iallocation1(0,8);

  generateDirections();

  print = 0;

}//--- TAngles ------------------------------------------

TAngles::~TAngles()
{
  freemem2(angles, 0,NoOfDirections,1,7);

  freemem1(c_pos, 0,2);
  freemem1(l_pos, 0,2);

  freemem1(i_weight, 0,2);
  freemem1(v_weight, 0,2);
  freeimem1(i_index, 0,2);
  freeimem1(v_index, 0,2);

  freemem1(texWeights, 0,8);
  freeimem1(illuPos, 0,8);
  freeimem1(viewPos, 0,8);
}//--- ~TAngles ------------------------------------------

void
TAngles::generateDirections()
{

  printf(" Measurement directions generation....\r");fflush(stdout);
  int theta = 15;
  int phi   = 0;
  int phi_increments[] = { 60, 30, 20, 18, 15 };

  angles[0][1] = 0.;
  angles[0][2] = 0.;          
  int index = 1;
  for (int i = 0; i < 5; ++i, theta+=15 )
    for (phi = 0; phi < 360 ; phi += phi_increments[i], ++index )
      {
        angles[index][1] = theta;
        angles[index][2] = phi;          
        angles2xyz(angles[index][1], angles[index][2], &angles[index][3], &angles[index][4], &angles[index][5]);
        //        printf("%d %f %f  %f %f %f\n",index, angles[index][1],angles[index][2],angles[index][3],angles[index][4],angles[index][5]);
      }
  printf(" Measurement directions generation....done\n");fflush(stdout);

}//--- generateDirections -----------------------------------------

void
TAngles::getViewPos(float vpos[])
{
  vpos[0] = c_pos[0];
  vpos[1] = c_pos[1];
  vpos[2] = c_pos[2];
  return;
};

void
TAngles::getIlluPos(float ipos[])
{
  ipos[0] = l_pos[0];
  ipos[1] = l_pos[1];
  ipos[2] = l_pos[2];
  return;
};

void
TAngles::setViewPos(float vpos[])
{
  c_pos[0] = vpos[0];
  c_pos[1] = vpos[1];
  c_pos[2] = vpos[2];
  return;
};

// Note that this assumes normalized values of T, B, N
int
TAngles::computeViewAngles(float V[], float T[], float B[], float N[],
			   float *theta_v, float *phi_v)
{
  // computing relative world coordinats of light and camera
  float view[3];
  vec_diff(c_pos, V, view);

  // transform view & illum vectors into local texture coordinates, i.e. tangent space
  float t_view[3];
  t_view[0] = T[0]*view[0] + T[1]*view[1] + T[2]*view[2];
  t_view[1] = B[0]*view[0] + B[1]*view[1] + B[2]*view[2];
  t_view[2] = N[0]*view[0] + N[1]*view[1] + N[2]*view[2];
  
  if ( (!vec_norm(t_view)) || (isnan_local(t_view[0])) ) {
    *theta_v = 0.f;
    *phi_v = 0.f;
    return -1;
  }

  *theta_v = acosf(t_view[2]);
  if (*theta_v > 91.f*M_PI/180.f) {
    *phi_v = 0.f;
    return -1;
  }

  if ((*theta_v > M_PI)&&(*theta_v < M_PI/2.f+.001f))
    *phi_v = 0.f;
  else
    // illum. and view angles computation
    *phi_v = atan2(t_view[1],t_view[0]);

  if (*phi_v < 0.f)
    *phi_v += M_PI*2.0f;
  if (*phi_v > M_PI*2.0f)
    *phi_v -= M_PI*2.0f;

  if (!( (*phi_v >= 0.f) && (*phi_v <= 2.0f*M_PI))) {
    printf("phi_v = %3.6f y=%f x=%f\n", *phi_v, t_view[1], t_view[0]);
  }
  assert( (*phi_v >= 0.f) && (*phi_v <= 2.0f*M_PI));

  //  if(this->print)
  //    printf("tv:%3.2f, pv:%3.2f\n",t_v,p_v);
  return 0;
}//--- computeViewAngles -------------------------------------------------

// Set illu position
void
TAngles::setIlluPos(float ipos[])
{
  l_pos[0] = ipos[0];
  l_pos[1] = ipos[1];
  l_pos[2] = ipos[2];
  return;
}//-- setLightPos------------------------------------

// Direction taken as point light source in infinity
void
TAngles::setIlluDir(float illudir[])
{
  l_pos[0] = illudir[0] * 1e10;
  l_pos[1] = illudir[1] * 1e10;
  l_pos[2] = illudir[2] * 1e10;
  return;
}//-- setIlluDir------------------------------------

void
TAngles::computeIlluAngles(float V[], float T[], float B[], float N[],
			   float *theta_i, float *phi_i)
{
  // computing relative world coordinates of light and camera
  float illu[3];
  vec_diff(l_pos, V, illu);
  vec_norm(illu); // normalize

  // transform illum vectors into local texture coordinates, i.e. tangent space
  float t_illuD[3];
  t_illuD[0] = T[0]*illu[0] + T[1]*illu[1] + T[2]*illu[2];
  t_illuD[1] = B[0]*illu[0] + B[1]*illu[1] + B[2]*illu[2];
  t_illuD[2] = N[0]*illu[0] + N[1]*illu[1] + N[2]*illu[2];
  vec_norm(t_illuD);
  *theta_i = acosf(t_illuD[2])*0.999999;
  assert(*theta_i >= 0.f);
  if (*theta_i > M_PI) {
    std::cout << "Problem: theta_i = " << *theta_i << std::endl;
    std::cout << "PI = " << M_PI << std::endl;
    std::cout << "t_illudD = " << t_illuD[0] << " " << t_illuD[1] << " " << t_illuD[2] << std::endl;
  }
  assert(*theta_i <= M_PI);

  // illum. and view angles computation
  *phi_i = atan2(t_illuD[1], t_illuD[0]);
  if(*phi_i < 0.f)
    *phi_i += 2.f*M_PI;
  if(*phi_i > 2.f*M_PI)
    *phi_i -= 2.f*M_PI;

  //  if(this->print)
  //    printf("ti:%3.2f, pi:%3.2f\n",t_i,p_i);

  // phi in range 0...2*M_PI, theta 0..M_PI
  return;
}//--- computeIlluAngles -------------------------------------------------------

int
TAngles::computeAnglesSphere(float *v, float *N)
{
  // computing relative world coordinats of light and camera
  float view[3], illu[3];
  vec_diff(c_pos, v, view);
  vec_norm(view); // normalize
  vec_diff(l_pos,v,illu);
  vec_norm(illu); // normalize

  // computing tangent space
  float T[3];
  float B[3];

  float alpha = atan2f(v[1], v[0]);
  B[0] =  sin(alpha);
  B[1] = -cos(alpha);
  B[2] = 0.f;
  vec_norm(B);
      
  vec_norm(N);
  vec_crossprod(B, N, T);
  vec_norm(T);

  // transform view & illum vectors into local texture coordinates, i.e. tangent space
  float t_view[3], t_illu[3];
  t_view[0] = T[0]*view[0] + T[1]*view[1] + T[2]*view[2];
  t_view[1] = B[0]*view[0] + B[1]*view[1] + B[2]*view[2];
  t_view[2] = N[0]*view[0] + N[1]*view[1] + N[2]*view[2];
  vec_norm(t_view);
  t_v = acosf(t_view[2]);
  if (t_v > M_PI/2.0f)
    return -1;

  t_illu[0] = T[0]*illu[0] + T[1]*illu[1] + T[2]*illu[2];
  t_illu[1] = B[0]*illu[0] + B[1]*illu[1] + B[2]*illu[2];
  t_illu[2] = N[0]*illu[0] + N[1]*illu[1] + N[2]*illu[2];
  vec_norm(t_illu);
  t_i = acosf(t_illu[2]);

  // illum. and view angles computation
  p_i = atan2(t_illu[1], t_illu[0]);
  if (p_i < 0)
    p_i += 2.f*M_PI;
  p_v = atan2(t_view[1], t_view[0]);
  if (p_v < 0)
    p_v += 2.f*M_PI;

  if(this->print)
    printf("ti:%3.2f, pi:%3.2f, tv:%3.2f, pv:%3.2f\n",t_i,p_i,t_v,p_v);

  return 0;
}//--- computeAnglesSphere ---------------------------------------------------

void
TAngles::getAngles(float *theta_i, float *phi_i, float *theta_v, float *phi_v)
{
  *theta_i = t_i;
  *phi_i   = p_i;
  *theta_v = t_v;
  *phi_v   = p_v;
}//--- getAngles ----------------------------------------------------------------

void
TAngles::computeTB(float **v, float **UV, float *T, float *B)
{
  // compute the 2.0 * surface area of the triangle
  float twoA = (UV[1][0] - UV[0][0]) * (UV[2][1] - UV[0][1]) +
               (UV[2][0] - UV[0][0]) * (UV[1][1] - UV[0][1]);
  assert(fabs(twoA) > 1e-12);
  twoA = 1.0 / twoA;
  for(int j=0;j<=2;j++) {
    T[j] = twoA * (v[0][j] * (UV[1][1] - UV[2][1]) + v[1][j] *
		   (UV[2][1] - UV[0][1]) + v[2][j] * (UV[0][1] - UV[1][1]));
    B[j] = twoA * (v[0][j] * (UV[2][0] - UV[1][0]) + v[1][j] *
		   (UV[0][0] - UV[2][0]) + v[2][j] * (UV[1][0] - UV[0][0]));
  } // for
  vec_norm(T);
  vec_norm(B);
  return;
}//--- computeTB ---------------------------------------------------



